import pytest

from knwl.config import get_config, merge_configs, reset_active_config, resolve_dict, set_active_config, merge_into_active_config
from knwl.config import resolve_reference, resolve_config, get_custom_config

pytestmark = pytest.mark.basic


def test_config_get():
    config = {"a": {"b": {"c": 1}}}
    assert get_config("a", "b", "c", config=config) == 1
    assert get_config("a", "b", "d", default=42, config=config) == 42
    assert get_config("a", "e", default=3, config=config) == 3
    assert get_config("x", default=None, config=config) is None
    assert get_config("a", "b", config=config) == {"c": 1}
    assert get_config("a", config=config) == {"b": {"c": 1}}
    assert get_config(config=config) == config
    assert (get_config("nonexistent", default="default_value", config=config) == "default_value")
    assert get_config("llm", "ollama", "model") == "qwen2.5:7b"
    assert get_config("llm", "ollama", "temperature") == 0.1
    assert get_config("llm", "ollama", "context_window") == 32768
    assert get_config("llm", "ollama", "caching_service") == "@/llm_caching/user"
    assert get_config("llm_caching", "user", "path") == "$/user/default/llm_cache.json"
    assert get_config("nonexistent", default="default_value") == "default_value"
    assert get_config("llm", "nonexistent", default={"key": "value"}) == {"key": "value"}
    assert get_config("storage", "documents", "nonexistent", default=123) == 123
    assert get_config("storage", "nonexistent", default={"a": 1}) == {"a": 1}
    assert get_config("nonexistent", default=None) is None
    assert (get_config("llm", "ollama", "model", override={"llm": {"ollama": {"model": "custom_model:1b"}}}, ) == "custom_model:1b")
    assert (get_config("llm", "ollama", "temperature", override={"llm": {"ollama": {"temperature": 0.56}}}, ) == 0.56)
    assert get_config("@/llm/ollama/model") == "qwen2.5:7b"
    assert get_config("@/a/b", override=config) == {"c": 1}
    assert get_config("@/a/b/", override=config) == {"c": 1}
    # should fetch the default if the reference is only the service name
    assert "class" in get_config("@/rag_store")


def test_config_merge():
    base = {"a": 1, "b": {"c": 2, "d": 3}}
    override = {"b": {"c": 20}, "e": 5}
    merged = {"a": 1, "b": {"c": 20, "d": 3}, "e": 5, }

    result = merge_configs(override, base)
    assert result == merged

    result = merge_configs({}, base)
    assert result == base
    result = merge_configs(override, {})
    assert result == override
    result = merge_configs({}, {})
    assert result == {}
    result = merge_configs(None, None)
    assert result == None

    with pytest.raises(ValueError):
        merge_configs("not_a_dict", base)
    with pytest.raises(ValueError):
        merge_configs(override, "not_a_dict")

    config = {"llm": {"openai": {"caching_service": "@/llm_caching/special"}}, "llm_caching": {"special": {"class": "A"}}, }
    result = get_config("llm", "openai", "caching_service", override=config)
    assert result == "@/llm_caching/special"


def test_resolve_reference():
    config = {"klm": {"default": "a", "a": {"u": 4, "v": 5, "w": 6}}, "abc": {"default": "b", "b": {"z": "@/klm"}, "c": {"y": "@/klm/a/u"}}, }
    result = resolve_reference("@/abc/b/z", override=config)
    assert result == {"u": 4, "v": 5, "w": 6}

    result = resolve_reference("@/abc/c/y", override=config)
    assert result == 4

    result = resolve_reference("@/klm", override=config)
    assert result == {"u": 4, "v": 5, "w": 6}

    result = resolve_reference("@/abc/c", override=config)
    assert result == {"y": 4}


def test_resolve_config():
    config = {"klm": {"default": "a", "a": {"u": 4, "v": 5, "w": 6}}, "abc": {"default": "b", "b": {"z": "@/klm"}, "c": {"y": "@/klm/a/u"}}, }
    result = resolve_config("abc", override=config)
    assert result == {"z": {"u": 4, "v": 5, "w": 6}}

    result = resolve_config("klm", override=config)
    assert result == {"u": 4, "v": 5, "w": 6}

    result = resolve_config("@/abc/c", override=config)
    assert result == {"y": 4}


def test_resolve_dict():
    d = {"a": {"default": "u", "u": {"k": 4}}}
    r = resolve_dict(d)
    assert r == {"a": {"k": 4}}

    d = {"a": {"default": "u", "u": {"k": "$/user/r"}}}
    r = resolve_dict(d)
    assert ".knwl" in r["a"]["k"]

    d = {"a": {"default": "u", "u": {"k": "@/b"}}, "b": {"default": "data", "data": {"d": 56}}, }
    r = resolve_dict(d, d)
    assert r["a"] == {"k": {"d": 56}}
    assert r["b"] == {"d": 56}


def test_active_config():
    dumb = {"a": 6}
    set_active_config(dumb)
    assert get_config("a") == 6
    assert get_config("@/llm") is None
    assert get_config("nonexistent", default=10) == 10
    assert get_config("a", override={"a": 12}) == 12
    reset_active_config()
    assert get_config("@/llm") is not None


def test_get_custom_config_namespace():
    # Test with default namespace
    config = get_custom_config(namespace="default")
    assert "$/user/default" in config["json"]["user_documents"]["path"]

    # Test with custom namespace
    config = get_custom_config(namespace="custom_space")
    assert "$/user/custom_space" in config["json"]["user_documents"]["path"]
    assert "$/user/custom_space" in config["json"]["user_chunks"]["path"]
    assert "$/user/custom_space" in config["graph"]["user"]["path"]
    assert "$/user/custom_space" in config["llm_caching"]["user"]["path"]

    # Test that original config is not modified
    original = get_config("json", "user_documents", "path")
    assert "$/user/default" in original


def test_get_custom_config_absolute_path(tmp_path):
    # Test with absolute path
    abs_path = str(tmp_path / "test_space")
    config = get_custom_config(namespace=abs_path)

    assert abs_path in config["json"]["user_documents"]["path"]
    assert abs_path in config["llm_caching"]["user"]["path"]
    assert abs_path in config["graph"]["user"]["path"]


def test_get_custom_config_llm_provider():
    # Test with LLM provider change
    config = get_custom_config(namespace="test", llm_provider="openai")
    assert config["llm"]["default"] == "openai"

    # Test with invalid provider
    with pytest.raises(ValueError, match="LLM provider.*not found"):
        get_custom_config(namespace="test", llm_provider="nonexistent")


def test_get_custom_config_llm_model():
    # Test with LLM model change
    config = get_custom_config(namespace="test", llm_provider="ollama", llm_model="custom:7b")
    assert config["llm"]["ollama"]["model"] == "custom:7b"

    # Test model change with openai
    config = get_custom_config(namespace="test", llm_provider="openai", llm_model="gpt-4")
    assert config["llm"]["openai"]["model"] == "gpt-4"


def test_get_custom_config_combined():
    # Test namespace + provider + model
    config = get_custom_config(namespace="myspace", llm_provider="anthropic", llm_model="claude-3")
    assert "$/user/myspace" in config["json"]["user_chunks"]["path"]
    assert config["llm"]["default"] == "anthropic"
    assert config["llm"]["anthropic"]["model"] == "claude-3"


def test_get_custom_config_empty_namespace():
    with pytest.raises(ValueError, match="namespace cannot be empty"):
        get_custom_config(namespace="")

    with pytest.raises(ValueError, match="namespace cannot be empty"):
        get_custom_config(namespace="   ")

    with pytest.raises(ValueError, match="namespace cannot be empty"):
        get_custom_config(namespace=None)


def test_get_custom_config_isolation():
    # Verify config isolation - changes don't affect active config
    original_default = get_config("llm", "default")

    config1 = get_custom_config(namespace="space1", llm_provider="openai")
    config2 = get_custom_config(namespace="space2", llm_provider="anthropic")

    assert config1["llm"]["default"] == "openai"
    assert config2["llm"]["default"] == "anthropic"
    assert get_config("llm", "default") == original_default


def test_extra():
    # Test with extra overrides
    extra = {"logging": {"enabled": False, "level": "ERROR", "path": "$/user/extra/knwl.log"}}
    config = get_custom_config(namespace="extra_space", override=extra)
    assert config["logging"]["enabled"] is False
    assert config["logging"]["level"] == "ERROR"
    assert "$/user/extra/knwl.log" in config["logging"]["path"]


def test_merge_into_active_config():
    reset_active_config()
    base_value = get_config("llm", "ollama", "temperature")
    override = {"llm": {"ollama": {"temperature": 0.75}}}
    merge_into_active_config(override)
    new_value = get_config("llm", "ollama", "temperature")
    assert new_value == 0.75

    found = get_config("a", "b")
    assert found is None
    merge_into_active_config({"a": {"b": 42}})
    found = get_config("a", "b")
    assert found == 42

    reset_active_config()
    assert get_config("a", "b") is None
    assert get_config("llm", "ollama", "temperature") == base_value
    with pytest.raises(ValueError):
        merge_into_active_config("not_a_dict")



from abc import ABC, abstractmethod

from knwl.framework_base import FrameworkBase
from knwl.models import  KnwlKeywords
from knwl.utils import hash_with_prefix

class KeywordsExtractionBase(FrameworkBase, ABC):
    """
    Abstract base class for keywords extraction implementations.

    Args:
        *args: Variable length argument list passed to parent class.
        **kwargs: Arbitrary keyword arguments passed to parent class.
            override: Optional configuration override dictionary.
    Methods:
        extract: Extract entities from text and return as dictionary.
        extract_records: Extract entities from text and return as list of records.
        extract_json: Extract entities from text and return as JSON dictionary.
    Note:
        This is an abstract base class and cannot be instantiated directly.
        All abstract methods must be implemented by subclasses.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

       

    @abstractmethod
    async def extract(
        self, text: str, chunk_id: str = None
    ) -> KnwlKeywords | None:
        ...




from .KnwlAnswer import KnwlAnswer
from .KnwlNode import KnwlNode
from .KnwlEdge import KnwlEdge
from .KnwlGraph import KnwlGraph
from .KnwlExtraction import KnwlExtraction
from .KnwlDocument import KnwlDocument
from .KnwlChunk import KnwlChunk
from .KnwlInput import KnwlInput
from .KnwlEntity import KnwlEntity
from .KnwlBlob import KnwlBlob
from .KnwlIngestion import KnwlIngestion
from .KnwlParams import KnwlParams
from .KnwlKeywords import KnwlKeywords
from .KnwlContext import KnwlContext
from .KnwlText import KnwlText
from .KnwlReference import KnwlReference
from .KnwlModel import KnwlModel

__all__ = ["KnwlNode", "KnwlEdge", "KnwlGraph", "KnwlExtraction", "KnwlDocument", "KnwlChunk", "KnwlInput",   "KnwlAnswer", "KnwlEntity", "KnwlBlob", "KnwlIngestion", "KnwlParams", "KnwlKeywords", "KnwlContext",  "KnwlText", "KnwlReference", "KnwlModel"]



from abc import ABC
from knwl.di import defaults
from knwl.models import KnwlBlob
from knwl.storage.blob_storage_base import BlobStorageBase
import os
import json
from pathlib import Path
from typing import Any, Optional

from knwl.utils import get_full_path


@defaults("blob", "file_system")
class FileStorage(BlobStorageBase, ABC):
    """
    Local file system implementation of blob storage.

    Stores blobs using a custom binary format that preserves both data and metadata
    in a single file:
    - First 4 bytes: metadata length (as 32-bit unsigned integer, big-endian)
    - Next N bytes: JSON metadata (UTF-8 encoded)
    - Remaining bytes: binary data
    """

    def __init__(self, base_path: Optional[str] = None):
        super().__init__()

        self.base_path = get_full_path(base_path or "$/data/files")

    async def upsert(self, blob: KnwlBlob) -> str | None:
        """
        Upsert a blob to a file with metadata.

        Stores blob in custom format: [metadata_length][metadata_json][binary_data]
        """
        file_path = os.path.join(self.base_path, blob.id)
        self.validate_blob(blob)

        # Prepare metadata (exclude data field)
        metadata = {
            "id": blob.id,
            "name": blob.name,
            "description": blob.description,
            "timestamp": blob.timestamp,
            "type_name": blob.type_name,
            "metadata": json.dumps(blob.metadata or {}),  # Store metadata as JSON string
        }
        metadata_json = json.dumps(metadata, ensure_ascii=False)
        metadata_bytes = metadata_json.encode("utf-8")

        # Write custom format: [4-byte length][metadata][data]
        with open(file_path, "wb") as f:
            # Write metadata length as 4-byte unsigned int (big-endian)
            f.write(len(metadata_bytes).to_bytes(4, byteorder="big"))
            # Write metadata
            f.write(metadata_bytes)
            # Write binary data
            f.write(blob.data)

        return blob.id

    async def get_by_id(self, id) -> KnwlBlob | None:
        """
        Get a blob by id from a file, reconstructing full object with metadata.
        """
        file_path = os.path.join(self.base_path, id)
        if not os.path.exists(file_path):
            return None

        with open(file_path, "rb") as f:
            # Read metadata length (4 bytes)
            metadata_length_bytes = f.read(4)
            if len(metadata_length_bytes) < 4:
                # Fallback for old format (just binary data)
                f.seek(0)
                data = f.read()
                return KnwlBlob(id=id, data=data)

            metadata_length = int.from_bytes(metadata_length_bytes, byteorder="big")

            # Read metadata
            metadata_bytes = f.read(metadata_length)
            metadata = json.loads(metadata_bytes.decode("utf-8"))

            # Read remaining binary data
            data = f.read()

        return KnwlBlob(
            id=metadata.get("id", id),
            name=metadata.get("name", ""),
            description=metadata.get("description", ""),
            timestamp=metadata.get("timestamp"),
            metadata=json.loads(metadata.get("metadata", "{}")) if metadata.get("metadata") else {},
            data=data,
        )

    async def delete_by_id(self, id: str) -> bool:
        """Delete a blob by id from a file."""
        file_path = os.path.join(self.base_path, id)
        if not os.path.exists(file_path):
            return False
        os.remove(file_path)
        return True

    async def count(self) -> int:
        """Count the number of blobs in the file storage."""
        return len(os.listdir(self.base_path))

    async def exists(self, id: str) -> bool:
        """Check if a blob exists by id in the file storage."""
        file_path = os.path.join(self.base_path, id)
        return os.path.exists(file_path)

    def validate_blob(self, blob: KnwlBlob) -> None:
        """Validate a blob before storage."""
        if (
            blob is not None
            and blob.id is not None
            and len(blob.id) > 0
            and blob.data is not None
            and len(blob.data) > 0
        ):
            return
        raise ValueError("Invalid blob provided for storage.")



import uuid
import pytest

from knwl import services
from knwl.format import print_knwl
from knwl.models import KnwlParams, KnwlDocument, KnwlContext, KnwlInput
from knwl.semantic.graph_rag.graph_rag import GraphRAG
from knwl.semantic.graph_rag.strategies.naive_strategy import NaiveGragStrategy
from knwl.utils import get_full_path
import os
pytestmark = pytest.mark.llm

from tests.library.collect import get_library_article


@pytest.mark.asyncio
async def test_naive_augmentation():
    content = await get_library_article("mathematics", "Topology")
    doc = KnwlDocument(content=content, id=f"{str(uuid.uuid4())}.txt")
    grag: GraphRAG = services.get_service("graph_rag")
    await grag.ingest(doc)
    input = KnwlInput(
        text="Explain the concept of homeomorphism in topology.",
        name="Test Query",
        description="A test query for topology concepts.",
        params=KnwlParams(strategy="naive"),
    )
    found = await grag.augment(input)
    print("")
    print_knwl(found, show_chunks=True, show_nodes=False, show_edges=False)

    """ 
    The above will render something like this:
    
╭───────────────────────────────── 🎯 Context ─────────────────────────────────╮
│                                                                              │
│                                                                              │
│  Question: Explain the concept of homeomorphism in topology.                 │
│                                                                              │
│                                                                              │
│ 📑 Chunks:                                                                   │
│                                                                              │
│ 📄[0] -to-one and onto, and if the inverse of the function is also           │
│ continuous, then the function is called a homeomorphism and the domain of    │
│ the function is...                                                           │
│                                                                              │
│                                                                              │
│ 📄[1] Topology (from the Greek words τόπος, 'place, location', and λόγος,    │
│ 'study') is the branch of mathematics concerned with the properties of a     │
│ geometric...                                                                 │
│                                                                              │
│                                                                              │
│ 📄[2] require distorting the space and affecting the curvature or volume.    │
│                                                                              │
│ Geometric topology                                                           │
│ Geometric topology is a branch of topology that primarily focu...            │
│                                                                              │
│                                                                              │
│ 📄[3] ic geometry. Donaldson, Jones, Witten, and Kontsevich have all won     │
│ Fields Medals for work related to topological field theory.                  │
│ The topological classif...                                                   │
│                                                                              │
│                                                                              │
│ 📄[4] en's theorem, covering spaces, and orbit spaces.)                      │
│ Wacław Sierpiński, General Topology, Dover Publications, 2000, ISBN          │
│ 0-486-41148-6                                                                │
│ Pickover, Clifford...                                                        │
│                                                                              │
│                                                                              │
╰───────────────────────── 5 chunks, 0 nodes, 0 edges ─────────────────────────╯
    """


@pytest.mark.asyncio
async def test_naive_augmentation():
    content = await get_library_article("mathematics", "Topology")
    doc = KnwlDocument(content=content, id=f"{str(uuid.uuid4())}.txt")
    grag: GraphRAG = services.get_service("graph_rag")
    await grag.ingest(doc)
    input = KnwlInput(
        text="Explain the concept of homeomorphism in topology.",
        name="Test Query",
        description="A test query for topology concepts.",
        params=KnwlParams(strategy="naive", return_chunks=True),
    )
    found = await grag.augment(input)
    print("")
    print_knwl(found, show_texts=True, show_nodes=False, show_edges=False)

    assert found is not None
    assert isinstance(found, KnwlContext)
    assert len(found.texts) > 0
    assert len(found.nodes) == 0
    assert len(found.edges) == 0
    assert found.input == input


@pytest.mark.asyncio
async def test_naive_strategy_initialization():
    """Test that NaiveGragStrategy can be initialized with a GraphRAG instance."""
    grag: GraphRAG = services.get_service("graph_rag")
    strategy = NaiveGragStrategy(grag)
    assert strategy.grag == grag


@pytest.mark.asyncio
async def test_naive_strategy_augment_with_no_results():
    """Test naive strategy when no chunks are found."""
    grag: GraphRAG = services.get_service("graph_rag")
    strategy = NaiveGragStrategy(grag)

    input = KnwlInput(
        text="A completely unrelated query that should find nothing",
        name="Empty Query",
        description="Test query that should return no results.",
        params=KnwlParams(strategy="naive", limit=5),
    )

    context = await strategy.augment(input)

    assert context is not None
    assert isinstance(context, KnwlContext)
    assert len(context.texts) == 0
    assert len(context.nodes) == 0
    assert len(context.edges) == 0
    assert context.input == input


@pytest.mark.asyncio
async def test_naive_strategy_augment_with_results():
    """Test naive strategy returns chunks in correct format."""
    content = await get_library_article("mathematics", "Topology")
    doc = KnwlDocument(content=content, id=f"{str(uuid.uuid4())}.txt")
    grag: GraphRAG = services.get_service("graph_rag")
    await grag.ingest(doc)

    strategy = NaiveGragStrategy(grag)
    input = KnwlInput(
        text="topology",
        name="Test Query",
        description="Simple test query.",
        params=KnwlParams(strategy="naive", top_k=3),
    )

    context = await strategy.augment(input)

    assert context is not None
    assert isinstance(context, KnwlContext)
    assert len(context.texts) <= 3
    assert len(context.nodes) == 0
    assert len(context.edges) == 0

    # Verify chunk structure
    for i, chunk in enumerate(context.texts):
        assert chunk.text is not None
        assert chunk.id is not None
        assert chunk.index == i


@pytest.mark.asyncio
async def test_naive_strategy_respects_limit_param():
    """Test that naive strategy respects the limit parameter."""
    content = await get_library_article("mathematics", "Topology")
    doc = KnwlDocument(content=content, id=f"{str(uuid.uuid4())}.txt")
    grag: GraphRAG = services.get_service("graph_rag")
    await grag.ingest(doc)

    strategy = NaiveGragStrategy(grag)

    # Test with limit=2
    input = KnwlInput(
        text="topology mathematics",
        name="Limited Query",
        description="Test query with limit.",
        params=KnwlParams(strategy="naive", top_k=2),
    )

    context = await strategy.augment(input)

    assert context is not None
    assert len(context.texts) <= 2



import pytest

from knwl.config import reset_config
from knwl.models.KnwlDocument import KnwlDocument
from knwl.semantic.rag.rag_store import RagStore
from tests.library.collect import get_library_article
pytestmark = pytest.mark.llm


@pytest.mark.asyncio
async def test_rag_store():
    reset_config()
    store = RagStore()
    article = await get_library_article("literature", "Jane Austen")
    doc = KnwlDocument(
        id="jane_austen",
        content=article,
        name="Jane Austen",
        description="An article about Jane Austen.",
    )
    id = await store.upsert(doc)
    assert await store.exists(id) is True

    # let's add another article unrelated to literature
    article = await get_library_article("arts", "Mona Lisa")
    doc = KnwlDocument(
        id="mona_lisa",
        content=article,
        name="Mona Lisa",
        description="An article about the Mona Lisa painting.",
    )
    await store.upsert(doc)

    # search for nearest chunks related to literature
    found = await store.nearest("epistolary novel")
    assert len(found) > 0
    ids = [chunk.origin_id for chunk in found]
    assert "jane_austen" in ids
    print(ids)
    
    # search for nearest chunks related to arts
    found = await store.nearest("Da Vinci")
    assert len(found) > 0
    ids = [chunk.origin_id for chunk in found]
    assert "mona_lisa" in ids
    print(ids)



from abc import ABC, abstractmethod

from knwl.framework_base import FrameworkBase
from knwl.models import KnwlEdge
from knwl.models.KnwlEdge import KnwlEdge
from knwl.models.KnwlExtraction import KnwlExtraction
from knwl.models.KnwlGraph import KnwlGraph
from knwl.models.KnwlNode import KnwlNode
from knwl.prompts import prompts


class GraphExtractionBase(FrameworkBase, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
       
    @abstractmethod
    async def extract(
        self, text: str, entities: list[str] = None, chunk_id: str = None
    ) -> KnwlExtraction | None:
        pass

    @abstractmethod
    async def extract_records(
        self, text: str, entities: list[str] = None
    ) -> list[list] | None:
        pass

    @abstractmethod
    async def extract_json(self, text: str, entities: list[str] = None) -> dict | None:
        pass

    @abstractmethod
    async def extract_graph(
        self, text: str, entities: list[str] = None, chunk_id: str = None
    ) -> KnwlGraph | None:
        pass

    @staticmethod
    def records_to_json(records: list[list]) -> dict:
        result = {
            "entities": [],
            "relationships": [],
            "keywords": [],
        }
        for rec in records:
            if rec[0] == "entity":
                try:
                    result["entities"].append(
                        {
                            "name": rec[1],
                            "type": rec[2],
                            "description": rec[3],
                        }
                    )
                except Exception as e:
                    from knwl.logging import log
                    log.error(f"Error parsing entity record: {rec}: \n{e}")
            elif rec[0] == "relationship":
                try:
                    result["relationships"].append(
                        {
                            "source": rec[1],
                            "target": rec[2],
                            "description": rec[3],
                            "types": (
                                [u.strip() for u in rec[4].split(",")]
                                if rec[4] is not None
                                else []
                            ),
                            "weight": float(rec[5]) if len(rec) > 5 and rec[5] else 1.0,
                        }
                    )
                except Exception as e:
                    from knwl.logging import log
                    log.error(f"Error parsing relationship record: {rec}: \n{e}")
            elif rec[0] == "content_keywords":
                result["keywords"].extend(rec[1].split(", "))
        #  make keywords unique
        result["keywords"] = list(set(result["keywords"]))
        return result

    @staticmethod
    def records_to_extraction(
        records: list[list], chunk_id: str = None
    ) -> KnwlExtraction:
        dic = GraphExtractionBase.records_to_json(records)
        if len(dic["entities"]) == 0:
            return KnwlExtraction(nodes={}, edges={}, keywords=[])
        nodes: dict[str, list[KnwlNode]] = {}
        edges: dict[str, list[KnwlEdge]] = {}

        node_map = {}  # map of node names to node ids
        for item in dic["entities"]:
            node = KnwlNode(
                name=item["name"],
                type=item["type"],
                description=item["description"],
                chunk_ids=[chunk_id] if chunk_id else [],
            )
            if node.name not in nodes:
                nodes[node.name] = [node]
            else:
                coll = nodes[node.name]
                found = [
                    e
                    for e in coll
                    if e.type == node.type and e.description == node.description
                ]
                if len(found) == 0:
                    coll.append(node)
            node_map[node.name] = node.id
        for item in dic["relationships"]:
            edge = KnwlEdge(
                source_id=item["source"],
                target_id=item["target"],
                description=item["description"],
                keywords=item["types"],
                weight=item["weight"],
                chunk_ids=[chunk_id] if chunk_id else [],
                type=item["types"][0] if len(item["types"]) > 0 else "Unknown",
            )
            # the edge key is the tuple of the source and target names, NOT the ids. Is corrected below
            edge_key = f"({edge.source_id},{edge.target_id})"
            if (edge.source_id, edge.target_id) not in edges:
                edges[edge_key] = [edge]
            else:
                coll = edges[edge_key]
                found = [
                    e
                    for e in coll
                    if e.description == edge.description and e.keywords == edge.keywords
                ]
                if len(found) == 0:
                    coll.append(edge)

        # the edge endpoints are the names and not the ids
        corrected_edges = {}
        for key in edges:
            for e in edges[key]:

                if e.source_id not in node_map or e.target_id not in node_map:
                    #  happens if the LLM creates edges to entities that are not in the graph
                    continue
                if key not in corrected_edges:
                    corrected_edges[key] = []
                source_id = node_map[e.source_id]
                target_id = node_map[e.target_id]
                corrected_edge = KnwlEdge(
                    source_id=source_id,
                    target_id=target_id,
                    description=e.description,
                    keywords=e.keywords,
                    weight=e.weight,
                    chunk_ids=e.chunk_ids,
                    type=e.type
                )
                corrected_edges[key].append(corrected_edge)
        return KnwlExtraction(
            nodes=nodes, edges=corrected_edges, keywords=dic["keywords"] or []
        )

    @staticmethod
    def extraction_to_graph(extraction: KnwlExtraction) -> KnwlGraph:
        nodes = []
        edges = []

        for name in extraction.nodes:
            for node in extraction.nodes[name]:
                nodes.append(node)
        for key in extraction.edges:
            for edge in extraction.edges[key]:
                edges.append(edge)
        return KnwlGraph(nodes=nodes, edges=edges, keywords=extraction.keywords or [])

    def get_extraction_prompt(self, text, entity_types=None):
        if self.extraction_mode == "fast":
            return prompts.extraction.fast_graph_extraction(text, entity_types)
        if self.extraction_mode == "full":
            return prompts.extraction.full_graph_extraction(text, entity_types)
        else:
            raise ValueError(f"Unknown extraction mode: {self.extraction_mode}")



# ============================================================================================
# Use VSCode Interactive Python for best experience but you can also run this script directly.
# See https://code.visualstudio.com/docs/python/jupyter-support-py
# ============================================================================================
# %%
from knwl import services, ChunkingBase
from knwl.format import print_knwl

config = {
    "chunking": {
        "tiktoken": {
            "chunk_size": 20,
            "chunk_overlap": 0,
        },
    }
}
s: ChunkingBase = services.get_service("chunking", override=config)
text = """In 1939, upon arriving late to his statistics course at the University of California, Berkeley, George Dantzig — a first-year graduate student — copied two problems off the blackboard, thinking they were a homework assignment. He found the homework “harder to do than usual,” he would later recount, and apologized to the professor for taking some extra days to complete it. A few weeks later, his professor told him that he had solved two famous open problems in statistics. Dantzig’s work would provide the basis for his doctoral dissertation and, decades later, inspiration for the film Good Will Hunting.

Dantzig received his doctorate in 1946, just after World War II, and he soon became a mathematical adviser to the newly formed U.S. Air Force. As with all modern wars, World War II’s outcome depended on the prudent allocation of limited resources. But unlike previous wars, this conflict was truly global in scale, and it was won in large part through sheer industrial might. The U.S. could simply produce more tanks, aircraft carriers and bombers than its enemies. Knowing this, the military was intensely interested in optimization problems — that is, how to strategically allocate limited resources in situations that could involve hundreds or thousands of variables.

The Air Force tasked Dantzig with figuring out new ways to solve optimization problems such as these. In response, he invented the simplex method, an algorithm that drew on some of the mathematical techniques he had developed while solving his blackboard problems almost a decade before.

Nearly 80 years later, the simplex method is still among the most widely used tools when a logistical or supply-chain decision needs to be made under complex constraints. It’s efficient and it works. “It has always run fast, and nobody’s seen it not be fast,” said Sophie Huiberts of the French National Center for Scientific Research (CNRS).

At the same time, there’s a curious property that has long cast a shadow over Dantzig’s method. In 1972, mathematicians proved that the time it takes to complete a task could rise exponentially with the number of constraints. So, no matter how fast the method may be in practice, theoretical analyses have consistently offered worst-case scenarios that imply it could take exponentially longer. For the simplex method, “our traditional tools for studying algorithms don’t work,” Huiberts said."""
result = await s.chunk(text, source_key="test")
for chunk in result:
    print("----chunk----")
    print(chunk)

# %%
# ============================================================================================
# Chunking via Knwl.
# In this case the chunking configuration is read from the knwl config.
# ============================================================================================
from knwl import Knwl
kg = Knwl()
print("===========Chunking Config======================")
print_knwl("@/chunking")
result = await kg.chunk(text)
for chunk in result:
    print("----chunk----")
    print(chunk)
# %%



import asyncio
import json
import os
import tempfile
from hashlib import md5
from unittest.mock import patch, MagicMock

import pytest

pytestmark = pytest.mark.basic

from knwl.config import resolve_dict
from knwl.models import KnwlAnswer
from knwl.models.KnwlChunk import KnwlChunk
from knwl.models.KnwlDocument import KnwlDocument
from knwl.models.KnwlEdge import KnwlEdge
from knwl.models.KnwlNode import KnwlNode
from knwl.utils import (
    _ensure_directories_exist,
    _resolve_reference_path,
    _resolve_special_prefixes,
    clean_str,
    pack_messages,
    split_string_by_multi_markers,
    hash_with_prefix,
)
from knwl.utils import hash_args, get_json_body, get_full_path, parse_llm_record
from knwl.utils import throttle


def test_valid_json_string():
    content = 'Some text before {"key": "value"} some text after'
    expected = '{"key": "value"}'
    result = get_json_body(content)
    assert result == expected


def test_no_json_string():
    content = "Some text without JSON"
    result = get_json_body(content)
    assert result is None


def test_empty_string():
    content = ""
    result = get_json_body(content)
    assert result is None


def test_multiple_json_strings():
    content = 'First JSON: {"key1": "value1"} Second JSON: {"key2": "value2"}'
    expected = '{"key1": "value1"}'
    result = get_json_body(content)
    assert result == expected


def test_json_string_with_nested_objects():
    content = 'Text with nested JSON: {"key": {"nested_key": "nested_value"}}'
    expected = '{"key": {"nested_key": "nested_value"}}'
    result = get_json_body(content)
    assert result == expected


def test_null_input():
    with pytest.raises(ValueError):
        get_json_body(None)


def test_compute_args_hash_single_arg():
    arg = "test"
    expected = md5(str((arg,)).encode()).hexdigest()
    result = hash_args(arg)
    assert result == expected


def test_compute_args_hash_multiple_args():
    args = ("test1", "test2", 123)
    expected = md5(str(args).encode()).hexdigest()
    result = hash_args(*args)
    assert result == expected


def test_compute_args_hash_no_args():
    expected = md5(str(()).encode()).hexdigest()
    result = hash_args()
    assert result == expected


def test_compute_args_hash_same_args_different_order():
    args1 = ("test1", "test2")
    args2 = ("test2", "test1")
    result1 = hash_args(*args1)
    result2 = hash_args(*args2)
    assert result1 != result2


def test_compute_args_hash_with_none():
    args = (None, "test")
    expected = md5(str(args).encode()).hexdigest()
    result = hash_args(*args)
    assert result == expected


def test_clean_str_html_escape():
    input_str = "Hello &amp; welcome!"
    expected = "Hello & welcome!"
    result = clean_str(input_str)
    assert result == expected


def test_clean_str_control_characters():
    input_str = "Hello\x00World\x1f!"
    expected = "HelloWorld!"
    result = clean_str(input_str)
    assert result == expected


def test_clean_str_non_string_input():
    input_data = 12345
    result = clean_str(input_data)
    assert result == input_data


def test_clean_str_empty_string():
    input_str = ""
    expected = ""
    result = clean_str(input_str)
    assert result == expected


def test_clean_str_whitespace():
    input_str = "   Hello World!   "
    expected = "Hello World!"
    result = clean_str(input_str)
    assert result == expected


def test_split_string_by_multi_markers_single_marker():
    content = "Hello, world! This is a test."
    markers = [","]
    expected = ["Hello", "world! This is a test."]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_split_string_by_multi_markers_multiple_markers():
    content = "Hello, world! This is a test."
    markers = [",", "!"]
    expected = ["Hello", "world", "This is a test."]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_split_string_by_multi_markers_no_markers():
    content = "Hello, world! This is a test."
    markers = []
    expected = ["Hello, world! This is a test."]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_split_string_by_multi_markers_empty_content():
    content = ""
    markers = [","]
    expected = [""]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_split_string_by_multi_markers_no_content():
    content = None
    markers = [","]
    with pytest.raises(TypeError):
        split_string_by_multi_markers(content, markers)


def test_split_string_by_multi_markers_whitespace():
    content = "   Hello, world!   "
    markers = [","]
    expected = ["Hello", "world!"]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_split_string_by_multi_markers_consecutive_markers():
    content = "Hello,, world!! This is a test."
    markers = [",", "!"]
    expected = ["Hello", "world", "This is a test."]
    result = split_string_by_multi_markers(content, markers)
    assert result == expected


def test_pack_messages_single_message():
    messages = ("Hello",)
    expected = [{"role": "user", "content": "Hello"}]
    result = pack_messages(*messages)
    assert result == expected


def test_pack_messages_multiple_messages():
    messages = ("Hello", "Hi", "How are you?")
    expected = [
        {"role": "user", "content": "Hello"},
        {"role": "assistant", "content": "Hi"},
        {"role": "user", "content": "How are you?"},
    ]
    result = pack_messages(*messages)
    assert result == expected


def test_pack_messages_no_messages():
    messages = ()
    expected = []
    result = pack_messages(*messages)
    assert result == expected


def test_pack_messages_alternating_roles():
    messages = ("Message 1", "Message 2", "Message 3", "Message 4")
    expected = [
        {"role": "user", "content": "Message 1"},
        {"role": "assistant", "content": "Message 2"},
        {"role": "user", "content": "Message 3"},
        {"role": "assistant", "content": "Message 4"},
    ]
    result = pack_messages(*messages)
    assert result == expected


def test_pack_messages_empty_string():
    messages = ("",)
    expected = [{"role": "user", "content": ""}]
    result = pack_messages(*messages)
    assert result == expected


@pytest.mark.asyncio
async def test_limit_async_func_call_single_call():
    @throttle(max_size=1)
    async def sample_func(x):
        return x

    result = await sample_func(5)
    assert result == 5


@pytest.mark.asyncio
async def test_limit_async_func_call_multiple_calls():
    @throttle(max_size=2)
    async def sample_func(x):
        await asyncio.sleep(0.1)
        return x

    tasks = [sample_func(i) for i in range(4)]
    results = await asyncio.gather(*tasks)
    assert results == [0, 1, 2, 3]


@pytest.mark.skip("Not relevant for the current implementation")
@pytest.mark.asyncio
async def test_limit_async_func_call_exceeding_limit():
    @throttle(max_size=2, waitting_time=0.01)
    async def sample_func(x):
        await asyncio.sleep(0.1)
        return x

    tasks = [sample_func(i) for i in range(100)]
    results = await asyncio.gather(*tasks)
    assert results == list(range(100))


@pytest.mark.asyncio
async def test_limit_async_func_call_with_waiting():
    @throttle(max_size=1, waitting_time=0.01)
    async def sample_func(x):
        await asyncio.sleep(0.1)
        return x

    tasks = [sample_func(i) for i in range(2)]
    results = await asyncio.gather(*tasks)
    assert results == [0, 1]


def test_chunk_class():
    with pytest.raises(ValueError):
        KnwlChunk(content="", tokens=2, index=0, origin_id="doc1")

    c = KnwlChunk(content="Hello", tokens=0, index=0, origin_id="doc1")
    # id is assigned based on the content
    assert "chunk|>" in c.id
    assert c.id == hash_with_prefix(c.content, prefix="chunk|>")


def test_document_class():
    with pytest.raises(ValueError):
        KnwlDocument(content="")

    c = KnwlDocument(content="Hello")
    # id is assigned based on the content
    assert "doc|>" in c.id
    assert c.id == KnwlDocument.hash_keys("Hello")


def test_node_class():
    with pytest.raises(ValueError):
        KnwlNode(name="")

    c = KnwlNode(name="Hello")
    # id is assigned based on the content
    assert "node|>" in c.id
    assert c.id == KnwlNode.hash_node(c)


def test_edge_class():
    with pytest.raises(ValueError):
        KnwlEdge(source_id="", target_id="")

    c = KnwlEdge(source_id="a", target_id="b")
    # id is assigned based on the content
    assert "edge|>" in c.id
    assert c.id == KnwlEdge.hash_edge(c)


def test_args_hash():
    a = KnwlAnswer(
        messages=[{"content": "Hello"}], llm_service="ollama", llm_model="qwen2.5:14b"
    )
    print(a.model_dump_json())


def test_get_full_path():
    p = get_full_path("knwl.llm")
    assert p.endswith("knwl.llm")
    assert ".knwl" in p

    p = get_full_path("a/llm", "$/tests")
    assert p.endswith("a/llm")
    assert "/tests" in p

    p = get_full_path("$/tests/abc")
    assert p.endswith("abc")
    assert "/tests" in p


def test_hash():
    assert (
        hash_with_prefix("hello", "pre|>")
        == "pre|>" + md5("hello".encode()).hexdigest()
    )
    assert hash_with_prefix("", "pre|>") == "pre|>" + md5("".encode()).hexdigest()
    assert (
        hash_with_prefix("hello", "hash|>")
        == "hash|>" + md5("hello".encode()).hexdigest()
    )
    assert hash_with_prefix("", "hash|>") == "hash|>" + md5("".encode()).hexdigest()

    assert hash_with_prefix(
        json.dumps({"a": 1, "b": 2, "c": {"x": 23, "y": 0.1}}), "json|>"
    ) == hash_with_prefix(
        json.dumps({"a": 1, "b": 2, "c": {"x": 23, "y": 0.1}}), "json|>"
    )
    assert hash_with_prefix(
        json.dumps({"a": 1, "b": 2, "c": {"x": 23, "y": 0.1}}), "json|>"
    ) != hash_with_prefix(
        json.dumps({"a": 2, "b": 2, "c": {"x": 23, "y": 0.1}}), "json|>"
    )


def test_split_string_by_multi_markers_edge_cases():
    assert split_string_by_multi_markers("", [","]) == [""]
    assert split_string_by_multi_markers("No markers here", [","]) == [
        "No markers here"
    ]
    assert split_string_by_multi_markers(",,,", [","]) == []
    assert split_string_by_multi_markers("Hello,,,World", [","]) == ["Hello", "World"]
    assert split_string_by_multi_markers("Hello!!!World???Test", ["!!!", "???"]) == [
        "Hello",
        "World",
        "Test",
    ]

    assert (
        split_string_by_multi_markers("a|b|c$c|d|e$<END>k|l|m$<END>", ["$", "<END>"])
    ) == ["a|b|c", "c|d|e", "k|l|m"]


def test_get_full_path_none_input():
    """Test that None input returns None"""
    result = get_full_path(None)
    assert result is None


def test_get_full_path_non_string_input():
    """Test that non-string input raises ValueError"""
    with pytest.raises(ValueError, match="File path must be a string"):
        get_full_path(123)


def test_get_full_path_basic_file():
    """Test basic file path resolution with default $/data reference"""
    result = get_full_path("test.txt")
    assert result is not None
    assert result.endswith("test.txt")
    assert ".knwl" in result


def test_get_full_path_data_prefix():
    """Test $/data prefix resolution"""
    result = get_full_path("$/data/test.txt")
    assert result is not None
    assert result.endswith("test.txt")
    assert "data" in result


def test_get_full_path_root_prefix():
    """Test $/root prefix resolution"""
    result = get_full_path("$/root/test.txt")
    assert result is not None
    assert result.endswith("test.txt")


def test_get_full_path_test_prefix():
    """Test $/tests prefix resolution"""
    result = get_full_path("$/tests/test.txt")
    assert result is not None
    assert result.endswith("test.txt")
    assert "tests" in result


def test_get_full_path_with_reference_path():
    """Test with explicit reference path"""
    with tempfile.TemporaryDirectory() as temp_dir:
        result = get_full_path("test.txt", temp_dir)
        assert result is not None
        assert result.startswith(temp_dir)
        assert result.endswith("test.txt")


def test_get_full_path_with_data_reference():
    """Test with $/data reference path"""
    result = get_full_path("test.txt", "$/data")
    assert result is not None
    assert result.endswith("test.txt")
    assert "data" in result


def test_get_full_path_with_root_reference():
    """Test with $/root reference path"""
    result = get_full_path("test.txt", "$/root")
    assert result is not None
    assert result.endswith("test.txt")


def test_get_full_path_with_test_reference():
    """Test with $/tests reference path"""
    result = get_full_path("test.txt", "$/tests")
    assert result is not None
    assert result.endswith("test.txt")
    assert "tests" in result


def test_get_full_path_non_absolute_reference():
    """Test that non-absolute reference path raises FileNotFoundError"""
    with pytest.raises(FileNotFoundError, match="Reference path must be absolute"):
        get_full_path("test.txt", "relative/path")


def test_get_full_path_invalid_reference_type():
    """Test that non-string reference path raises ValueError"""
    with pytest.raises(ValueError, match="Reference path must be a string"):
        get_full_path("test.txt", 123)


def test_get_full_path_create_dirs_false():
    """Test with create_dirs=False"""
    result = get_full_path("test.txt", create_dirs=False)
    assert result is not None
    assert result.endswith("test.txt")


def test_get_full_path_nested_path():
    """Test with nested directory structure"""
    result = get_full_path("$/tests/subdir/nested/test.txt")
    assert result is not None
    assert result.endswith("subdir/nested/test.txt")


def test_get_full_path_absolute_result():
    """Test that result is always absolute path"""
    result = get_full_path("test.txt")
    assert result is not None
    assert os.path.isabs(result)


def test_resolve_special_prefixes():
    """Test _resolve_special_prefixes helper function"""
    # Test $/tests prefix
    path, ref = _resolve_special_prefixes("$/tests/file.txt")
    assert path == "./file.txt"
    assert ref == "$/tests"

    # Test $/data prefix
    path, ref = _resolve_special_prefixes("$/data/file.txt")
    assert path == "./file.txt"
    assert ref == "$/data"

    # Test $/root prefix
    path, ref = _resolve_special_prefixes("$/root/file.txt")
    assert path == "./file.txt"
    assert ref == "$/root"

    # Test no prefix
    path, ref = _resolve_special_prefixes("file.txt")
    assert path == "file.txt"
    assert ref is None


def test_resolve_reference_path():
    """Test _resolve_reference_path helper function"""
    # Test $/data resolution
    result = _resolve_reference_path("$/data", create_dirs=False)
    assert result is not None
    assert "data" in result

    # Test $/root resolution
    result = _resolve_reference_path("$/root", create_dirs=False)
    assert result is not None

    # Test $/tests resolution
    result = _resolve_reference_path("$/tests", create_dirs=False)
    assert result is not None
    assert "tests" in result


def test_resolve_reference_path_absolute():
    """Test _resolve_reference_path with absolute path"""
    with tempfile.TemporaryDirectory() as temp_dir:
        result = _resolve_reference_path(temp_dir, create_dirs=False)
        assert result == temp_dir


def test_resolve_reference_path_non_absolute():
    """Test _resolve_reference_path with non-absolute path raises error"""
    with pytest.raises(FileNotFoundError, match="Reference path must be absolute"):
        _resolve_reference_path("relative/path", create_dirs=False)


def test_ensure_directories_exist_file():
    """Test _ensure_directories_exist for file path"""
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = os.path.join(temp_dir, "subdir", "test.txt")
        _ensure_directories_exist(file_path)
        assert os.path.exists(os.path.dirname(file_path))


def test_ensure_directories_exist_directory():
    """Test _ensure_directories_exist for directory path"""
    with tempfile.TemporaryDirectory() as temp_dir:
        dir_path = os.path.join(temp_dir, "new_directory")
        _ensure_directories_exist(dir_path)
        assert os.path.exists(dir_path)


def test_ensure_directories_exist_error():
    """Test _ensure_directories_exist with creation error"""
    with patch("os.makedirs") as mock_makedirs:
        mock_makedirs.side_effect = OSError("Permission denied")

        with pytest.raises(OSError, match="Failed to create directories"):
            _ensure_directories_exist("/some/path/test.txt")


def test_special_dirs():
    p = get_full_path("$/root/xyz")
    print(p)
    assert p.endswith("xyz") and "knwl" in p and p.startswith("/")
    if os.path.exists(p):
        os.rmdir(p)

    p = get_full_path("$/user/abc")
    print(p)
    assert p.endswith("abc") and ".knwl" in p and p.startswith(os.path.expanduser("~"))
    if os.path.exists(p):
        os.rmdir(p)

    p = get_full_path("$/data/xyz")
    print(p)
    assert p.endswith("xyz") and p.startswith("/")
    if os.path.exists(p):
        os.rmdir(p)

    p = get_full_path("$/tests/xyz")
    print(p)
    assert p.endswith("xyz") and p.startswith("/")
    if os.path.exists(p):
        os.rmdir(p)


def test_parse_llm_record():
    rec = "(component1, component2, component3)"
    result = parse_llm_record(rec, delimiter=",")
    assert result == ["component1", "component2", "component3"]

    rec = "No parentheses here"
    result = parse_llm_record(rec, delimiter=",")
    assert result is None

    rec = ""
    result = parse_llm_record(rec, delimiter=",")
    assert result is None

    rec = None
    result = parse_llm_record(rec, delimiter=",")
    assert result is None

    rec_missing_end_parentheses = "(entity<|>Catherine Thomson Hogarth<|>person<|>Catherine Thomson Hogarth was the daughter of George Hogarth and became Charles Dickens's wife after a one-year engagement."
    result = parse_llm_record(rec_missing_end_parentheses, delimiter="<|>")
    assert result == [
        "entity",
        "Catherine Thomson Hogarth",
        "person",
        "Catherine Thomson Hogarth was the daughter of George Hogarth and became Charles Dickens's wife after a one-year engagement.",
    ]



"""
Markdown formatter for Knwl models.

This module provides Markdown output for Knwl models, useful for
documentation, reports, and integration with static site generators.
"""

import json
from typing import Any, Dict, List
from pydantic import BaseModel

from knwl.format.formatter_base import FormatterBase, ModelFormatter, get_registry
from knwl.models.KnwlGraph import KnwlGraph


class MarkdownFormatter(FormatterBase):
    """
    Formatter for Markdown output.

    Creates GitHub-flavored Markdown with tables and code blocks.
    """

    def __init__(self):
        """Initialize the Markdown formatter."""
        self._registry = get_registry()
        self._indent_level = 0

    def format(self, obj: Any, **options) -> str:
        """
        Format an object as Markdown.

        Args:
            obj: The object to format
            **options: Formatting options

        Returns:
            Markdown string
        """
        if obj is None:
            return "_None_"
        if isinstance(obj, str):
            if obj.strip().startswith("@/"):
                from knwl.config import resolve_config

                return json.dumps(resolve_config(obj.strip()), indent=2)
            else:
                return f"To render a configuration, use the '@/service/variant' syntax. For example, `@/vector/memory', '@/llm/openai' or simply '@/' for the whole config."
        # Handle Pydantic models
        if isinstance(obj, BaseModel):
            formatter_class = self._registry.get_formatter(type(obj), "markdown")
            if formatter_class:
                formatter = formatter_class()
                return formatter.format(obj, self, **options)
            else:
                return self._format_default_model(obj, **options)

        # Handle lists
        if isinstance(obj, list):
            return self._format_list(obj, **options)

        # Handle dicts
        if isinstance(obj, dict):
            return self._format_dict(obj, **options)

        # Handle primitives
        return self._format_primitive(obj)

    def render(self, obj: Any, **options) -> None:
        """
        Render Markdown to stdout (or save to file if specified).

        Args:
            obj: The object to render
            **options: Rendering options (can include 'output_file')
        """
        md_output = self.format(obj, **options)

        # Add frontmatter if requested
        if options.get("add_frontmatter", False):
            title = options.get("title", "Knwl Output")
            md_output = self._add_frontmatter(md_output, title)

        output_file = options.get("output_file")
        if output_file:
            with open(output_file, "w", encoding="utf-8") as f:
                f.write(md_output)
        else:
            print(md_output)

    def _format_default_model(self, model: BaseModel, **options) -> str:
        """Format a Pydantic model with default Markdown styling."""
        title = options.get("title", model.__class__.__name__)
        level = options.get("level", 2)

        lines = []
        lines.append(f"{'#' * level} {title}\n")

        # Create table
        lines.append("| Field | Value |")
        lines.append("|-------|-------|")

        for field_name, field_value in model.model_dump().items():
            value_str = self._format_value_for_table(field_value)
            lines.append(f"| **{field_name}** | {value_str} |")

        return "\n".join(lines)

    def _format_list(self, items: List, **options) -> str:
        """Format a list as Markdown."""
        if not items:
            return "_Empty list_"

        lines = []
        for item in items:
            if isinstance(item, BaseModel):
                # For models, create a sub-section
                item_md = self.format(item, level=options.get("level", 2) + 1)
                lines.append(f"- {item_md}")
            else:
                lines.append(f"- {self._escape(str(item))}")

        return "\n".join(lines)

    def _format_dict(self, data: Dict, **options) -> str:
        """Format a dictionary as Markdown table."""
        if not data:
            return "_Empty dictionary_"

        lines = []
        lines.append("| Key | Value |")
        lines.append("|-----|-------|")

        for key, value in data.items():
            value_str = self._format_value_for_table(value)
            lines.append(f"| `{self._escape(str(key))}` | {value_str} |")

        return "\n".join(lines)

    def _format_primitive(self, value: Any) -> str:
        """Format primitive values as Markdown."""
        if isinstance(value, bool):
            return f"**{value}**" if value else f"_{value}_"
        elif isinstance(value, (int, float)):
            return f"`{value}`"
        elif isinstance(value, str):
            return self._escape(value)
        else:
            return self._escape(str(value))

    def _format_value_for_table(self, value: Any) -> str:
        """Format a value for display in a Markdown table cell."""
        if value is None:
            return "_None_"
        elif isinstance(value, bool):
            return f"**{value}**" if value else f"_{value}_"
        elif isinstance(value, list):
            return f"_[{len(value)} items]_"
        elif isinstance(value, dict):
            return f"_{{{len(value)} keys}}_"
        elif isinstance(value, BaseModel):
            return f"`{value.__class__.__name__}`"
        else:
            str_value = str(value)
            if len(str_value) > 80:
                str_value = str_value[:80] + "..."
            return self._escape(str_value)

    def _escape(self, text: str) -> str:
        """Escape special Markdown characters."""
        # Escape pipe characters for tables
        text = text.replace("|", "\\|")
        return text

    def _add_frontmatter(self, content: str, title: str) -> str:
        """Add YAML frontmatter to the Markdown."""
        from datetime import datetime

        frontmatter = f"""---
title: {title}
date: {datetime.now().isoformat()}
generator: Knwl Framework
---

"""
        return frontmatter + content

    def create_heading(self, text: str, level: int = 2) -> str:
        """Create a Markdown heading."""
        return f"{'#' * level} {text}\n"

    def create_table(self, headers: list[str], rows: list[list[str]]) -> str:
        """Create a Markdown table."""
        lines = []

        # Header
        lines.append("| " + " | ".join(headers) + " |")
        lines.append("|" + "|".join(["---" for _ in headers]) + "|")

        # Rows
        for row in rows:
            lines.append(
                "| " + " | ".join([self._escape(str(cell)) for cell in row]) + " |"
            )

        return "\n".join(lines)

    def create_code_block(self, code: str, language: str = "") -> str:
        """Create a Markdown code block."""
        return f"```{language}\n{code}\n```"

    def render_mermaid(self, graph: "KnwlGraph", **options) -> None:
        """
        Render a KnwlGraph as a Mermaid diagram in Markdown.

        Args:
            graph: The KnwlGraph object to render
            **options: Additional rendering options
        """
        mermaid_lines = ["```mermaid", "graph TD"]

        # for node in graph.nodes:
        #     mermaid_lines.append(f'    {node.id}["{self._escape(node.name)}"]')

        for edge in graph.edges:
            source_node = graph.get_node_by_id(edge.source_id)
            target_node = graph.get_node_by_id(edge.target_id)
            if source_node and target_node:
                mermaid_lines.append(
                    f"    {source_node.name} -->|{self._escape(edge.type)}| {target_node.name}"
                )

        mermaid_lines.append("```")

        mermaid_code = "\n".join(mermaid_lines)
        output_file = options.get("output_file")
        if output_file:
            with open(output_file, "w", encoding="utf-8") as f:
                f.write(mermaid_code)
        else:
            print(mermaid_code)



from knwl.models.KnwlInput import KnwlInput
from knwl.utils import hash_with_prefix
from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Optional
from datetime import datetime


class KnwlDocument(BaseModel):
    """
    A class representing a source document.

    Attributes:
        content (str): The content of the source document.
        id (str): A unique identifier for the source document. Defaults to a new UUID.
        timestamp (str): The timestamp when the source document was created. Defaults to the current time in ISO format.
        typeName (str): The type name of the source document. Defaults to "KnwlDocument".
        name (str): The name of the source document. Defaults to an empty string.
        description (str): A description of the source document. Defaults to an empty string.
    """

    content: str = Field(..., description="The content of the document")
    id: Optional[str] = Field(
        default=None, description="Unique identifier for the document"
    )
    timestamp: str = Field(
        default_factory=lambda: datetime.now().isoformat(),
        description="Creation timestamp",
    )
    description: Optional[str] = Field(default="", description="Document description")
    name: Optional[str] = Field(default="", description="Document name")
    type_name: str = Field(
        default="KnwlDocument",
        frozen=True,
        description="The type name of this class, for serialization purposes.",
    )
    # todo: have ontology in a separate ontology store


    @field_validator("content")
    def content_not_empty(cls, v):
        if v is None or len(str.strip(v)) == 0:
            raise ValueError("Content of a KnwlDocument cannot be None or empty.")
        return v

    @model_validator(mode="after")
    def set_id(self) -> "KnwlDocument":
        if self.id is None and self.content is not None and len(str.strip(self.content)) > 0:
            object.__setattr__(
                self, "id", self.hash_keys(self.content, self.name, self.description)
            )
        return self

    @staticmethod
    def from_input(input: KnwlInput):
        return KnwlDocument(
            content=input.text, name=input.name, description=input.description, id=input.id
        )

    @staticmethod
    def hash_keys(content: str, name: str = None, description: str = None) -> str:
        return hash_with_prefix(
            content + " " + (name or "") + " " + (description or ""), prefix="doc|>"
        )



import pytest

from knwl.llm.json_llm_cache import JsonLLMCache
from knwl.models import KnwlAnswer
pytestmark = pytest.mark.basic
cache = JsonLLMCache()


@pytest.mark.asyncio
async def test_llm_cache():
    d = {
        "messages": [{"role": "user", "content": "hello"}],
        "llm_model": "qwen2.5:14b",
        "llm_service": "ollama",
        "answer": "world",
        "timing": 20.0,
    }
    a = KnwlAnswer(**d)
    id = await cache.upsert(a)
    assert await cache.is_in_cache(
        **{
            "messages": [{"role": "user", "content": "hello"}],
            "llm_model": "qwen2.5:14b",
            "llm_service": "ollama",
        }
    )
    cached_item = await cache.get_by_id(id)
    assert cached_item.from_cache
    assert cached_item.messages == [{"role": "user", "content": "hello"}]
    assert await cache.is_in_cache("hello", "ollama", "qwen2.5:14b")
    assert not await cache.is_in_cache("hello again", "ollama", "qwen2.5:14b")
    assert not await cache.is_in_cache("hello", "ollama", "another_model")
    await cache.delete(a)
    assert not await cache.is_in_cache("hello", "ollama", "qwen2.5:14b")
    with pytest.raises(ValueError):
        await cache.upsert(None)
    with pytest.raises(ValueError):
        await cache.upsert("not a KnwlAnswer")



from typing import Literal

from pydantic import BaseModel, Field

AugmentationStrategy = Literal["local", "global", "hybrid", "naive", "none", "self"]


class KnwlParams(BaseModel):
    """
    Parameters tuning graph RAG augmentation.

    Attributes:
        strategy (AugmentationStrategy): The query strategy to use - local, global, hybrid, self/none or naive.
        top_k (int): Number of top-k items to retrieve (entities in local mode, relationships in global mode).
        return_chunks (bool): Whether to return the chunks. If not, only the chunk Id's are returned and downstream services need to fetch the chunk data separately.
    """

    strategy: AugmentationStrategy = Field(
        default="local",
        description="The query strategy to use - local, global, hybrid, self or naive.",
    )

    top_k: int = Field(
        default=5,
        description="Number of top-k items to retrieve (entities in local mode, relationships in global mode).",
    )

    return_chunks: bool = Field(
        default=True,
        description="Whether to return the chunk text data or just the chunk Id's.",
    )



from typing import List, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from knwl.utils import hash_with_prefix


class KnwlEdge(BaseModel):
    """
    Represents a relation between atoms of knowledge.

    Minimum required fields are source_id, target_id and type.
    Contrary to most models this one is not immutable because it complicates the grag algorithms too much.

    Attributes:
        source_id (str): The Id of the source node.
        target_id (str): The Id of the target node.
        chunk_ids (list[str]): The IDs of the chunks.
        weight (float): The weight of the edge.
        description (Optional[str]): A description of the edge.
        keywords (list[str]): Keywords associated with the edge.
        type_name (str): The type name of the edge, default is "KnwlEdge".
        id (str): The unique identifier of the edge, default is a new UUID.
    """

    degree: Optional[int] = Field(
        default=None, description="The degree of the edge in the graph."
    )
    source_id: str = Field(description="The Id of the source node.")
    target_id: str = Field(description="The Id of the target node.")
    source_name: Optional[str] = Field(
        default=None, description="The name of the source node."
    )
    target_name: Optional[str] = Field(
        default=None, description="The name of the target node."
    )
    type: Optional[str] = Field(
        default="Unknown",
        description="The type of the relation. In a property modeled graph this should be an ontology class.",
    )
    type_name: str = Field(
        default="KnwlEdge",
        frozen=True,
        description="The type name of the edge for (de)serialization purposes.",
    )
    id: Optional[str] = Field(
        default=None,
        description="The unique identifier of the node, automatically generated from the required fields.",
    )
    chunk_ids: Optional[list[str]] = Field(
        default_factory=list,
        description="The chunk identifiers associated with this edge.",
    )
    keywords: Optional[list[str]] = Field(
        default_factory=list,
        description="Keywords associated with the edge. These can be used as types or labels in a property graph. Note that the names of the keywords should ideally be from an ontology.",
    )

    description: Optional[str] = Field(
        default="", description="A description of the edge."
    )
    weight: Optional[float] = Field(
        default=1.0,
        description="The weight of the edge. This can be used to represent the strength or importance of the relationship. This is given by domain experts or derived from data extraction.",
    )
    index: Optional[int] = Field(
        default=0, description="The index of the edge within the parent list, if any."
    )
    data: Optional[dict] = Field(
        default_factory=dict,
        description="Additional data associated with the knowledge node.",
    )

    @staticmethod
    def hash_edge(e: "KnwlEdge") -> str:
        return hash_with_prefix(
            e.source_id + " " + e.target_id + " " + e.type,
            prefix="edge|>",
        )

    def has_data(self, key: str) -> bool:
        """
        Check if the KnwlNode has any additional data.

        Returns:
            bool: True if the data dictionary is not empty, False otherwise.
        """

        return key in self.data

    def get_data(self, key: str):
        """
        Get additional data associated with the KnwlNode.

        Returns:
            The value associated with the key in the data dictionary, or None if the key does not exist.
        """

        return self.data.get(key, None)

    @field_validator("source_id")
    @classmethod
    def validate_source_id(cls, v):
        if v is None or len(str(v).strip()) == 0:
            raise ValueError("Source Id of a KnwlEdge cannot be None or empty.")
        return v

    @field_validator("target_id")
    @classmethod
    def validate_target_id(cls, v):
        if v is None or len(str(v).strip()) == 0:
            raise ValueError("Target Id of a KnwlEdge cannot be None or empty.")
        return v

    @field_validator("type")
    @classmethod
    def validate_type(cls, v):
        if v is None or len(str(v).strip()) == 0:
            raise ValueError("Type of a KnwlEdge cannot be None or empty.")
        return v.strip().replace("<", "").replace(">", "") # sanitize accidents from LLMs

    @model_validator(mode="after")
    def update_id(self):
        # Note that using only source and target is not enough to ensure uniqueness
        if self.id is None:
            object.__setattr__(self, "id", KnwlEdge.hash_edge(self))
        if self.type is None or len(self.type.strip()) == 0:
            if self.keywords and len(self.keywords) > 0:
                object.__setattr__(self, "type", self.keywords[0])
            else:
                object.__setattr__(self, "type", "Unknown")
        return self

    @field_validator("data", mode="before")
    @classmethod
    def parse_data(cls, v):
        """Parse JSON string to dict if needed."""
        if v is not None and isinstance(v, str):
            import json

            try:
                return json.loads(v)
            except json.JSONDecodeError as e:
                raise ValueError(f"Invalid JSON string for data field: {e}")
        return v

    @staticmethod
    def other_endpoint(edge: "KnwlEdge", node_id: str) -> str:
        if edge.source_id == node_id:
            return edge.target_id
        elif edge.target_id == node_id:
            return edge.source_id
        else:
            raise ValueError(f"Node {node_id} is not an endpoint of edge {edge.id}")

    def to_text(self) -> str:
        return f"""
Type: {self.type}
Source: {self.source_name}
Target: {self.target_name}
Description: {self.description}
Keywords: {', '.join(self.keywords) if self.keywords else 'None'}
    """


from typing import Any
import uuid
from knwl.storage.blob_storage_base import BlobStorageBase
from knwl.storage.graph_base import GraphStorageBase
from knwl.storage.kv_storage_base import KeyValueStorageBase
from knwl.storage.storage_base import StorageBase
from knwl.storage.vector_storage_base import VectorStorageBase


class StorageAdapter:
    """
    This sits between Knwl input formats and the diverse storage interfaces.
    It attempts to store whatever you throw at it in the appropriate storage backend.

    """

    @staticmethod
    async def upsert(obj: Any, storage: StorageBase | list[StorageBase]) -> str:
        """
        Upserts an object into the given storage backend(s).
        """
        if obj is None:
            raise ValueError("StorageAdapter: cannot upsert None object")
        if not isinstance(storage, list):
            storage = [storage]

        for store in storage:
            if store is None:
                continue
            if isinstance(store, KeyValueStorageBase):
                return await store.upsert(StorageAdapter.to_key_value(obj))
            elif isinstance(store, BlobStorageBase):
                pass
            elif isinstance(store, VectorStorageBase):
                pass
            elif isinstance(store, GraphStorageBase):
                pass
            else:
                raise ValueError(
                    f"StorageAdapter: unsupported upsert storage type: {type(store)}"
                )

    @staticmethod
    def to_key_value(obj: Any) -> dict[str, Any]:
        """
        Converts an object to a key-value representation for key-value storage.
        """
        if obj is None:
            return None
        if isinstance(obj, dict):
            return obj

        elif hasattr(obj, "id"):
            # single object with id attribute
            return {str(obj.id): obj.model_dump(mode="json")}
        elif hasattr(obj, "model_dump"):
            # single object with model_dump method
            return {
                str(getattr(obj, "id", str(uuid.uuid4()))): obj.model_dump(mode="json")
            }
        elif hasattr(obj, "dict"):
            # single object with dict method
            return {str(getattr(obj, "id", str(uuid.uuid4()))): obj.dict()}

        elif isinstance(obj, (float, int, bool, str, bytes)):
            return {str(uuid.uuid4()): obj}
        else:
            raise ValueError(
                f"StorageAdapter: cannot convert object of type {type(obj)} to dict"
            )



from unittest.mock import AsyncMock, MagicMock, patch
import pytest

from knwl.models import KnwlDocument, KnwlEdge
from knwl.models.KnwlChunk import KnwlChunk
from knwl.models.KnwlText import KnwlText
from knwl.models.KnwlNode import KnwlNode
from knwl.semantic.graph_rag.strategies.strategy_base import GragStrategyBase


class DummyStrategy(GragStrategyBase):
    async def augment(self, input):
        return None


@pytest.mark.asyncio
async def test_semantic_nodes():
    grag = MagicMock()
    n1 = KnwlNode(
        id="node1",
        name="Node 1",
        type="A",
        description="Test node 1",
        index=1,
        degree=5,
    )
    n2 = KnwlNode(
        id="node2",
        name="Node 2",
        type="B",
        description="Test node 2",
        index=0,
        degree=10,
    )

    # mock the nearest_nodes method
    grag.nearest_nodes = AsyncMock(return_value=[n1, n2])
    # mock the get_node_by_id method
    grag.get_node_by_id = AsyncMock(side_effect=lambda x: n1 if x == "node1" else n2)
    # mock the node_degree method
    grag.node_degree = AsyncMock(
        side_effect=lambda x: n1.degree if x == "node1" else n2.degree
    )
    strategy = DummyStrategy(grag)
    nodes = await strategy.semantic_node_search("test query")
    assert len(nodes) == 2
    assert isinstance(nodes[0], KnwlNode)
    assert nodes[0].id == "node2"  # node with higher degree should be first

    grag.node_degree = AsyncMock(side_effect=lambda x: None)
    nodes = await strategy.semantic_node_search("test query")
    assert len(nodes) == 2  # node degree missing, but still returns nodes

    grag.nearest_nodes = AsyncMock(return_value=None)
    assert await strategy.semantic_node_search("test query") == None

    grag.nearest_nodes = AsyncMock(return_value=[])
    assert await strategy.semantic_node_search("test query") == None


@pytest.mark.asyncio
async def test_semantic_edges():
    grag = MagicMock()
    e1 = KnwlEdge(
        source_id="node1",
        target_id="node2",
        type="related_to",
        description="Edge 1",
        weight=2.0,
        degree=4,
    )
    e2 = KnwlEdge(
        source_id="node2",
        target_id="node3",
        type="known_for",
        description="Edge 2",
        weight=3.0,
        degree=6,
    )
    grag.nearest_edges = AsyncMock(return_value=[e1, e2])
    grag.edge_degree = AsyncMock(
        side_effect=lambda id: (
            e1.degree if id == "edge1" else e2.degree
        )
    )
    grag.get_node_by_id = AsyncMock(
        side_effect=lambda x: KnwlNode(
            id=x,
            name=f"node{x[-1]}",
            type="A",
            description=f"Test node {x[-1]}",
            index=int(x[-1]),
        )
    )

    strategy = DummyStrategy(grag)
    edges = await strategy.semantic_edge_search("test query")
    assert len(edges) == 2
    assert isinstance(edges[0], KnwlEdge)
    assert edges[0].source_name == "node2"  # edge with higher degree should be first
    assert edges[0].target_name == "node3"


@pytest.mark.asyncio
async def test_nodes_from_edges():
    grag = MagicMock()
    n1 = KnwlNode(
        id="node1",
        name="Node 1",
        type="A",
        description="Test node 1",
        index=1,
        degree=5,
    )
    n2 = KnwlNode(
        id="node2",
        name="Node 2",
        type="B",
        description="Test node 2",
        index=0,
        degree=10,
    )
    n3 = KnwlNode(
        id="node3",
        name="Node 3",
        type="C",
        description="Test node 3",
        index=0,
        degree=2,
    )
    e1 = KnwlEdge(
        source_id="node1",
        target_id="node2",
        type="related_to",
        description="Edge 1",
        weight=2.0,
    )
    e2 = KnwlEdge(
        source_id="node2",
        target_id="node3",
        type="related_to",
        description="Edge 2",
        weight=2.4,
    )
    node_map = {
        "node1": n1,
        "node2": n2,
        "node3": n3,
    }
    grag.get_node_by_id = AsyncMock(side_effect=lambda x: node_map.get(x))
    degrees = {"node1": 5, "node2": 10, "node3": 2}
    grag.node_degree = AsyncMock(side_effect=lambda x: degrees.get(x, 0))

    strategy = DummyStrategy(grag)
    nodes = await strategy.nodes_from_edges([e1])
    assert len(nodes) == 2
    assert isinstance(nodes[0], KnwlNode)
    assert nodes[0].id == "node2"  # node with higher degree should be first


@pytest.mark.asyncio
async def test_edges_from_nodes():
    grag = MagicMock()
    n1 = KnwlNode(
        id="node1",
        name="Node 1",
        type="A",
        description="Test node 1",
        index=1,
        degree=5,
    )
    n2 = KnwlNode(
        id="node2",
        name="Node 2",
        type="B",
        description="Test node 2",
        index=0,
        degree=10,
    )
    e1 = KnwlEdge(
        id="edge1",
        source_id="node1",
        target_id="node2",
        type="related_to",
        description="Edge 1",
        weight=2.0,
        degree=4,
    )
    e2 = KnwlEdge(
        id="edge2",
        source_id="node2",
        target_id="node3",
        type="known_for",
        description="Edge 2",
        weight=3.0,
        degree=6,
    )
    grag.get_attached_edges = AsyncMock(return_value=[e1, e2])
    grag.assign_edge_degrees = AsyncMock()
    grag.get_semantic_endpoints = AsyncMock(
        return_value={
            "edge1": ("node1", "node2"),
            "edge2": ("node2", "node3"),
        }
    )
    grag.get_node_by_id = AsyncMock(
        side_effect=lambda x: KnwlNode(
            id=x,
            name=f"node{x[-1]}",
            type="A",
            description=f"Test node {x[-1]}",
            index=int(x[-1]),
        )
    )

    strategy = DummyStrategy(grag)
    edges = await strategy.edges_from_nodes([n1, n2])
    assert len(edges) == 2
    assert isinstance(edges[0], KnwlEdge)
    assert edges[0].source_name == "node2"  # edge with higher degree should be first
    assert edges[0].target_name == "node3"
    assert edges[0].index == 0


@pytest.mark.asyncio
async def test_chunk_stats():
    grag = MagicMock()

    nodes = [
        KnwlNode(
            id="node1",
            name="Node 1",
            type="A",
            description="Test node 1",
            index=1,
            degree=5,
            chunk_ids=["chunk1", "chunk2", "chunk3"],
        ),
        KnwlNode(
            id="node2",
            name="Node 2",
            type="B",
            description="Test node 2",
            index=0,
            degree=10,
            chunk_ids=["chunk2", "chunk3", "chunk4"],
        ),
        KnwlNode(
            id="node3",
            name="Node 3",
            type="B",
            description="Test node 3",
            index=0,
            degree=7,
            chunk_ids=[],
        ),
    ]
    edge1 = KnwlEdge(
        id="edge1",
        source_id="node1",
        target_id="node2",
        type="related_to",
        description="Edge 1",
        weight=2.0,
    )
    edge2 = KnwlEdge(
        id="edge2",
        source_id="node2",
        target_id="node3",
        type="related_to",
        description="Edge 2",
        weight=2.4,
    )
    grag.get_attached_edges = AsyncMock(return_value=[edge1, edge2])
    grag.get_node_by_id = AsyncMock(
        side_effect=lambda x: next((n for n in nodes if n.id == x), None)
    )
    grag.get_semantic_endpoints = AsyncMock(
        return_value={
            "edge1": ("node1", "node2"),
            "edge2": ("node2", "node3"),
        }
    )
    strategy = DummyStrategy(grag)
    stats = await strategy.chunk_stats_from_nodes(nodes)
    assert stats == {
        "chunk2": 1,  # appears once in edge1 at both endpoints
        "chunk3": 1,  # appears once in edge2 at both endpoints
        "chunk1": 0,  # not common in any edge
        "chunk4": 0,
    }


@pytest.mark.asyncio
async def test_unique_chunks():
    nodes = [
        KnwlNode(
            id="node1",
            name="Node 1",
            type="A",
            description="Test node 1",
            index=1,
            degree=5,
            chunk_ids=["chunk1", "chunk2", "chunk3"],
        ),
        KnwlNode(
            id="node2",
            name="Node 2",
            type="B",
            description="Test node 2",
            index=0,
            degree=10,
            chunk_ids=["chunk2", "chunk3", "chunk4"],
        ),
        KnwlNode(
            id="node2",
            name="Node 2",
            type="B",
            description="Test node 2",
            index=0,
            degree=10,
            chunk_ids=[],
        ),
    ]
    ids = DummyStrategy.unique_chunk_ids(nodes)
    assert set(ids) == {"chunk1", "chunk2", "chunk3", "chunk4"}


@pytest.mark.asyncio
async def test_text_from_nodes():
    grag = MagicMock()

    chunk_map = {
        "chunk1": KnwlChunk(id="chunk1", content="This is chunk 1."),
        "chunk2": KnwlChunk(id="chunk2", content="This is chunk 2."),
        "chunk3": KnwlChunk(id="chunk3", content="This is chunk 3."),
    }

    grag.get_chunk_by_id = AsyncMock(side_effect=lambda x: chunk_map.get(x))

    strategy = DummyStrategy(grag)
    strategy.chunk_stats_from_nodes = AsyncMock(
        return_value={
            "chunk1": 1,
            "chunk2": 5,
            "chunk3": 2,
        }
    )
    texts = await strategy.texts_from_nodes(
        [{}, {}], params=MagicMock(return_chunks=True)
    )
    assert len(texts) == 3
    assert texts[0].id == "chunk2"  # highest count
    assert texts[0].text == "This is chunk 2."


@pytest.mark.asyncio
async def test_references():
    grag = MagicMock()
    texts = [
        KnwlText(id="t1", text="Text 1", index=2, origin_id="chunk1"),
        KnwlText(id="t2", text="Text 2", index=5, origin_id="chunk2"),
    ]
    grag.get_chunk_by_id = AsyncMock(
        side_effect=lambda x: (
            {"id": x, "content": f"Content for {x}"}
            if x in ["chunk1", "chunk2"]
            else None
        )
    )
    grag.get_source_by_id = AsyncMock(
        side_effect=lambda x: (
            KnwlDocument(
                id=f"Source of {x}",
                content=f"Content for {x}",
                name=f"Source for {x}",
                description="Desc",
                timestamp="now",
            )
            if x in ["chunk1", "chunk2"]
            else None
        )
    )
    strategy = DummyStrategy(grag)
    references = await strategy.references_from_texts(texts)
    assert len(references) == 2
    assert references[0].content == "Content for chunk1"
    assert references[0].document_id == "Source of chunk1"


@pytest.mark.asyncio
async def test_references_from_chunks():
    grag = MagicMock()
    chunk_map = {
        "chunk1": KnwlChunk(id="chunk1", content="This is chunk 1.", origin_id="Origin 1"),
        "chunk2": KnwlChunk(id="chunk2", content="This is chunk 2.", origin_id="Origin 2"),
        "chunk3": KnwlChunk(id="chunk3", content="This is chunk 3.", origin_id="Origin 3"),
    }

    grag.get_chunk_by_id = AsyncMock(side_effect=lambda x: chunk_map.get(x))
    grag.get_source_by_id = AsyncMock(
        side_effect=lambda x: (
            KnwlDocument(
                id=f"Document of {x}",
                content=f"Content of {x}",
                name=f"Source for {x}",
                description="Desc",
                timestamp="now",
            )
            if x in ["Origin 1", "Origin 2", "Origin 3"]
            else None
        )
    )
    strategy = DummyStrategy(grag)
    references = await strategy.references_from_texts(chunk_map.values())
    assert len(references) == 3
    assert references[0].content == "Content of Origin 1"



import os.path

from knwl.models.KnwlContext import KnwlContext
from knwl.prompts.prompt_constants import PromptConstants

current_dir = os.path.dirname(os.path.abspath(__file__))


class RagPrompts:
    def __init__(self):
        self._self_rag_prompt = None

    def self_rag(self, question: str) -> str:
        if self._self_rag_prompt is None:
            with open(os.path.join(current_dir, "templates", "self_rag.txt"), "r") as f:
                self._self_rag_prompt = f.read()
        return self._self_rag_prompt.format(
            text=question,
            record_delimiter=PromptConstants.DEFAULT_RECORD_DELIMITER,
            completion_delimiter=PromptConstants.DEFAULT_COMPLETION_DELIMITER,
        )

    def grag_ask(self, question: str, augmentation: KnwlContext) -> str:
        with open(os.path.join(current_dir, "templates", "grag_ask.txt"), "r") as f:
            grag_ask_prompt = f.read()

        return grag_ask_prompt.format(
            input=question,
            text=question,
            nodes="\n ".join([n.to_text() for n in augmentation.nodes]),
            edges="\n\n ".join([e.to_text() for e in augmentation.edges]),
            texts=[t.text for t in augmentation.texts],
        )



from abc import ABC, abstractmethod
from typing import List

from knwl.framework_base import FrameworkBase
from knwl.models import KnwlAnswer


class LLMCacheBase(FrameworkBase):
    """
    Abstract base class for LLM caching functionality.
    """

    @abstractmethod
    async def is_in_cache(self, messages: str | list[str] | list[dict], llm_service: str, llm_model: str) -> bool:
        """Check if the given messages are cached for the specified LLM service and model."""
        pass

    @abstractmethod
    async def get(self, messages: str | list[str | list[dict]], llm_service: str, llm_model: str) -> KnwlAnswer | None:
        """Retrieve cached LLM answer for the given messages, service, and model."""
        pass

    @abstractmethod
    async def get_all_ids(self) -> list[str]:
        """Get all cached item IDs."""
        pass

    @abstractmethod
    async def save(self):
        """Save the cache to persistent storage."""
        pass

    @abstractmethod
    async def clear_cache(self):
        """Clear all cached items."""
        pass

    @abstractmethod
    async def get_by_id(self, id: str) -> KnwlAnswer | None:
        """Retrieve cached item by its ID."""
        pass

    @abstractmethod
    async def get_by_ids(self, ids: list[str], fields=None):
        """Retrieve multiple cached items by their IDs."""
        pass

    @abstractmethod
    async def filter_new_ids(self, data: list[str]) -> set[str]:
        """Filter out IDs that are already cached."""
        pass

    @abstractmethod
    async def upsert(self, a: KnwlAnswer) -> str:
        """Insert or update a cached LLM answer."""
        pass

    @abstractmethod
    async def delete_by_id(self, id: str):
        """Delete cached item by its ID."""
        pass

    @abstractmethod
    async def delete(self, a: KnwlAnswer):
        """Delete cached LLM answer."""
        pass



import pytest

from knwl.semantic.rag.chunk_store import ChunkStore
from faker import Faker
fake =Faker()

@pytest.mark.asyncio
async def test_crud():
    store = ChunkStore()
    text = fake.text()
    chunk_id = await store.upsert(text)
    assert chunk_id is not None
    chunk = await store.get_by_id(chunk_id)
    assert chunk is not None
    assert chunk.content == text.strip()
    assert await store.exists(chunk_id) is True
    await store.delete_by_id(chunk_id)
    assert await store.exists(chunk_id) is False



"""
Terminal formatting for Knwl models using Rich.
"""

from knwl.format.terminal.rich_formatter import RichFormatter, RichTheme
from knwl.format.terminal.model_formatters import (
    KnwlNodeTerminalFormatter,
    KnwlEdgeTerminalFormatter,
    KnwlGraphTerminalFormatter,
    KnwlDocumentTerminalFormatter,
    KnwlChunkTerminalFormatter,
    KnwlEntityTerminalFormatter,
    KnwlExtractionTerminalFormatter,
    KnwlContextTerminalFormatter,
    KnwlResponseTerminalFormatter,
)

__all__ = [
    "RichFormatter",
    "RichTheme",
    "KnwlNodeTerminalFormatter",
    "KnwlEdgeTerminalFormatter",
    "KnwlGraphTerminalFormatter",
    "KnwlDocumentTerminalFormatter",
    "KnwlChunkTerminalFormatter",
    "KnwlEntityTerminalFormatter",
    "KnwlExtractionTerminalFormatter",
    "KnwlContextTerminalFormatter",
    "KnwlResponseTerminalFormatter",
]



import asyncio
import os

import pytest
import random
from knwl.models.KnwlDocument import KnwlDocument
from knwl.storage.json_storage import JsonStorage
from knwl.utils import random_name, load_json
from faker import Faker
pytestmark = pytest.mark.basic

fake = Faker()


@pytest.fixture
def test_store():
    return JsonStorage("memory")


@pytest.mark.asyncio
async def test_delete(test_store):
    await test_store.upsert({"key1": {"value": "data1"}})
    await test_store.delete_by_id("key1")
    data = await test_store.get_by_id("key1")
    assert data is None


@pytest.mark.asyncio
async def test_all_keys(test_store):
    assert test_store.save_to_disk is False
    await test_store.clear()
    await test_store.upsert({"key1": {"value": "data1"}})
    await test_store.upsert({"key2": {"value": "data2"}})
    keys = await test_store.get_all_ids()
    assert set(keys) == {"key1", "key2"}


@pytest.mark.asyncio
async def test_save_somewhere():

    storage = JsonStorage(f"$/tests/{fake.word()}.json")
    data = {"key1": {"value": "data1"}}
    await storage.upsert(data)
    await storage.save()
    file_path = storage.path
    assert os.path.exists(file_path)
    # remove the JSON file and the directory again
    os.remove(file_path)


@pytest.mark.asyncio
async def test_get_by_id(test_store):
    await test_store.upsert({"key1": {"value": "data1"}})
    data = await test_store.get_by_id("key1")
    assert data == {"value": "data1"}


@pytest.mark.asyncio
async def test_get_by_ids(test_store):
    await test_store.upsert({"key1": {"value": "data1"}})
    await test_store.upsert({"key2": {"value": "data2"}})
    data = await test_store.get_by_ids(["key1", "key2"])
    assert data == [{"value": "data1"}, {"value": "data2"}]


@pytest.mark.asyncio
async def test_filter_keys(test_store):
    k1 = fake.word()
    k2 = fake.word()
    await test_store.upsert({k1: {"value": "data1"}})
    filtered_keys = await test_store.filter_new_ids([k1, k2])
    assert filtered_keys == {k2}


@pytest.mark.asyncio
async def test_upsert(test_store):
    data = {"key1": {"value": "data1"}}
    await test_store.upsert(data)
    stored_data = await test_store.get_by_id("key1")
    assert stored_data == data["key1"]


@pytest.mark.asyncio
async def test_drop(test_store):
    await test_store.upsert({"key1": {"value": "data1"}})
    await test_store.clear()
    keys = await test_store.get_all_ids()
    assert keys == []


@pytest.mark.asyncio
async def test_save_source(test_store):
    id = random_name()
    source = KnwlDocument(id=id, content="This is a test document.")
    await test_store.upsert({id: source})
    print()
    found = await test_store.get_by_id(id)
    print(found)
    await test_store.save()
    assert found == source.model_dump(mode="json")


@pytest.mark.asyncio
async def test_save():
    store = JsonStorage("test")
    await store.clear_cache()
    await store.clear()

    data = {"key1": {"value": "data1"}}
    await store.upsert(data)
    await store.save()
    file_path = store.path
    await asyncio.sleep(1)  # give os a moment to write the file
    assert os.path.exists(file_path)
    data = load_json(file_path)
    assert data == {"key1": {"value": "data1"}}
    await store.clear_cache()
    assert not os.path.exists(file_path)


@pytest.mark.asyncio
async def test_memory_only():
    store = JsonStorage("memory")
    await store.clear_cache()
    await store.clear()

    data = {"key1": {"value": "data1"}}
    await store.upsert(data)
    await store.save()
    assert store.path is None
    found = await store.get_by_id("key1")
    assert found == {"value": "data1"}
    await store.clear_cache()


@pytest.mark.asyncio
async def test_polymorphic(test_store):
    id = await test_store.upsert("simple string data")
    assert await test_store.exists(id)
    found = await test_store.get_by_id(id)
    assert found == "simple string data"
    val = random.random()
    id2 = await test_store.upsert(val)
    assert await test_store.exists(id2)
    found2 = await test_store.get_by_id(id2)
    assert found2 == val


@pytest.mark.asyncio
async def test_get_by_metadata(test_store):
    await test_store.upsert({"key1": {"value": "data1", "meta": "a"}})
    await test_store.upsert({"key2": {"value": "data2", "meta": "b"}})
    await test_store.upsert({"key3": {"value": "data3", "meta": "a"}})
    found = await test_store.get_by_metadata(meta="a")
    assert len(found) == 2
    values = [item["value"] for item in found]
    assert "data1" in values
    assert "data3" in values



from knwl.di import defaults
from knwl.extraction.graph_extraction_base import GraphExtractionBase
from knwl.llm.llm_base import LLMBase
from knwl.logging import log
from knwl.models.KnwlExtraction import KnwlExtraction
from knwl.models.KnwlGraph import KnwlGraph
from knwl.prompts import prompts
from knwl.utils import answer_to_records

continue_prompt = prompts.extraction.iterate_entity_extraction
if_loop_prompt = prompts.extraction.glean_break


@defaults("graph_extraction")
class BasicGraphExtraction(GraphExtractionBase):
    def __init__(self, llm: LLMBase = None, mode: str = "full"):
        super().__init__()
        if llm is None:
            raise ValueError("BasicGraphExtraction: LLM instance must be provided.")
        if not isinstance(llm, LLMBase):
            raise TypeError("BasicGraphExtraction: llm must be an instance of LLMBase.")
        self._llm = llm
        self.extraction_mode = mode

    @property
    def llm(self) -> LLMBase:
        return self._llm

    

    async def extract_records(self, text: str, entities: list[str] = None) -> list[list] | None:
        if not text or text.strip() == "":
            return None
        extraction_prompt = self.get_extraction_prompt(text, entity_types=entities)
        found = await self.llm.ask(question=extraction_prompt, key=text, category="graph-extraction")
        if not found or found.answer.strip() == "":
            return None
        return answer_to_records(found.answer)

    async def extract_json(self, text: str, entities: list[str] = None) -> dict | None:
        records = await self.extract_records(text, entities)
        if not records:
            return None
        return GraphExtractionBase.records_to_json(records)

    async def extract(self, text: str, entities: list[str] = None, chunk_id: str = None) -> KnwlExtraction | None:
        records = await self.extract_records(text, entities)
        if not records:
            return None
        return GraphExtractionBase.records_to_extraction(records, chunk_id)

    async def extract_graph(self, text: str, entities: list[str] = None, chunk_id: str = None) -> KnwlGraph | None:
        extraction = await self.extract(text, entities, chunk_id=chunk_id)
        if not extraction:
            return None
        return GraphExtractionBase.extraction_to_graph(extraction)

    


from datetime import datetime
from knwl.knwl import Knwl
from knwl.models import KnwlGraph
from knwl.models.KnwlIngestion import KnwlIngestion
from knwl.models.KnwlInput import KnwlInput
from knwl.models.KnwlParams import AugmentationStrategy
import time
import os
import csv


class Benchmark:

    def __init__(
        self,
        models: dict[str, list[str]] = {},
        facts: dict[str, str] = {},
        strategy: AugmentationStrategy = "local",
    ):

        self.models = models
        self.facts = facts
        self.strategy = strategy
        self.namespace = os.path.join(os.path.dirname(__file__), "knwl_data")

        self.ingest_file_path = os.path.join(
            self.ensure_results_dir(),
            f"ingest_{datetime.now().strftime('%Y%m%d_%H%M%S')}.csv",
        )
        if os.path.exists(self.ingest_file_path):
            os.remove(self.ingest_file_path)

    def ensure_results_dir(self):
        current_dir = os.path.dirname(os.path.abspath(__file__))

        dir_path = os.path.join(
            current_dir,
            "results",
        )
        os.makedirs(dir_path, exist_ok=True)
        return dir_path

    async def run(self) -> None:
        await self.ingest()
        await self.ask()

    async def ask(self) -> None:
        pass

    async def ingest(self) -> None:
        fieldnames = [
            "key",
            "provider",
            "model",
            "failed",
            "node_count",
            "edge_count",
            "latency",
            "error",
        ]
        with open(self.ingest_file_path, "a", newline="") as csvfile:

            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

            writer.writeheader()
            for provider, model_list in self.models.items():
                for model in model_list:
                    knwl = Knwl(namespace=self.namespace, llm=provider, model=model)
                    print(f"\n\n============ {provider}/{model} =================")

                    for key, text in self.facts.items():
                        input = KnwlInput(text=text, name=key)
                        node_count = 0
                        edge_count = 0
                        latency = 0.0
                        failed = False
                        try:
                            start_time = time.perf_counter()
                            result: KnwlGraph = await knwl.ingest(input)
                            end_time = time.perf_counter()
                            if result is None:
                                latency = end_time - start_time
                                node_count = 0
                                edge_count = 0
                            else:
                                latency = round(end_time - start_time, 2)
                                node_count = len(result.nodes)
                                edge_count = len(result.edges)
                            failed = False
                            error = " "
                        except Exception as e:
                            latency = 0.0
                            node_count = 0
                            edge_count = 0
                            failed = True
                            error = str(e)
                            print(error)
                        if failed:
                            print(f"{key}: {error}")
                        else:
                            print(f"{key}: {node_count}, {edge_count}, {latency}")
                        # Write row to CSV
                        writer.writerow(
                            {
                                "key": key,
                                "provider": provider,
                                "model": model,
                                "failed": failed,
                                "node_count": node_count,
                                "edge_count": edge_count,
                                "latency": latency,
                                "error": error,
                            }
                        )



import os
from typing import Any, cast

from pydantic import BaseModel

from knwl import KnwlModel
from knwl.di import defaults
from knwl.logging import log
from knwl.storage.kv_storage_base import KeyValueStorageBase
from knwl.storage.storage_adapter import StorageAdapter
from knwl.utils import load_json, write_json, get_full_path


@defaults("json")
class JsonStorage(KeyValueStorageBase):
    """
    Basic JSON storage implementation with everything in a single file.
    Note that this stores a dictionary of objects, where each object must have a unique 'id' field.
    It doesn't allow arrays of objects at the top level.
    """

    def __init__(self, path: str = "memory", save_to_disk: bool = True):
        """
        Initialize the JsonSingleStorage instance.
        """
        super().__init__()

        self._path = path
        self._save_to_disk = save_to_disk

        try:
            if (
                self._path == "memory"
                or self._path == "none"
                or self._path == "false"
                or self._path is False
                or (isinstance(self._path, bool) and not self._path)
                or (isinstance(self._save_to_disk, bool) and not self._save_to_disk)
            ):
                self._save_to_disk = False
                self._path = None
            else:
                if not self._path.endswith(".json"):
                    log.warn(
                        f"Json storage path '{self._path}' does not end with .json, appending .json"
                    )
                    self._path += ".json"
                self._path = get_full_path(self._path)
                self._save_to_disk = True

            if self._save_to_disk:
                if os.path.exists(self._path) and not os.path.isdir(self._path):
                    self.data = load_json(self._path) or {}
                    if len(self.data) > 0:
                        log(f"Loaded '{self._path}' JSON with {len(self.data)} items.")
                else:
                    self.data = {}
                    self._path = self._path
            else:
                self.data = {}
                self._path = None
        except Exception as e:
            log(e)

    @property
    def path(self):
        return self._path

    @property
    def save_to_disk(self):
        return self._save_to_disk

    async def get_all_ids(self) -> list[str]:
        """
        Get all keys in the storage.
        Returns: A list of id's (strings).
        """
        return list(self.data.keys())

    async def save(self):
        """
        Save the current data to the JSON file if storage is enabled.
        Returns: None
        """
        if self._save_to_disk:
            write_json(self.data, self._path)

    async def clear_cache(self):
        """
        Asynchronously removes the file if it exists.

        This method checks if the file specified by the instance variable `_file_name` exists.
        If the file exists, it removes the file.

        Raises:
            OSError: If an error occurs during file removal.
        """

        if self._save_to_disk and os.path.exists(self._path):
            os.remove(self._path)

    async def get_by_id(self, id):
        """
        Get a single item by its Id.
        Args:
            id: a string representing the Id of the item to retrieve.
        Returns:

        """
        return self.data.get(id, None)

    async def get_by_ids(self, ids, fields=None):
        """
        Get multiple items by their Ids.
        Args:
            ids: a list of strings representing the Ids of the items to retrieve.
            fields: an optional list of strings representing the fields to include in the result. If None, all fields are included.
        Returns: A list of items corresponding to the provided Ids. If an Id does not exist, None is returned in its place.
        """
        if fields is None:
            return [self.data.get(id, None) for id in ids]
        return [
            (
                {k: v for k, v in self.data[id].items() if k in fields}
                if self.data.get(id, None)
                else None
            )
            for id in ids
        ]

    async def filter_new_ids(self, data: list[str]) -> set[str]:
        """
        Filter out IDs that are already present in the storage, returning only unknown Id's.

        Args:
             data: A list of Id's to filter.
        Returns: A set of Id's that are not present in the storage.
        """
        return set([s for s in data if s not in self.data])

    async def upsert(self, obj: Any) -> str:
        """
        Upsert data into the JSON storage.

        This method updates existing data or inserts new data into the storage.

        Args:
            data (Any): Preferably a dictionary containing a key-value pairs to upsert, but it tries hard to accept anything.

        """
        if obj is None:
            raise ValueError("JsonStorage: cannot upsert None object")
        item = StorageAdapter.to_key_value(obj)
        if len(item.keys()) > 1:
            raise ValueError(
                "JsonStorage: can only upsert single items with unique ids"
            )
        if item is None:
            raise ValueError("JsonStorage: cannot upsert None item after conversion")
        left_data = {k: v for k, v in item.items() if k not in self.data}
        k = None
        for k in left_data:
            if isinstance(left_data[k], BaseModel):
                left_data[k] = left_data[k].model_dump(mode="json")

            elif isinstance(left_data[k], BaseModel):
                left_data[k] = left_data[k].model_dump(mode="json")
            else:
                left_data[k] = cast(dict, left_data[k])
            self.data.update(left_data)
        await self.save()
        return k  # the last one is the only one

    async def clear(self):
        """
        Clear all data from the storage and delete the file if it exists.
        """
        self.data = {}
        if self._save_to_disk and os.path.exists(self._path):
            os.remove(self._path)

    async def count(self):
        """
        Count the number of items in the storage.
        Returns: The number of items in the storage.
        """
        return len(self.data)

    async def delete_by_id(self, id: str):
        """
        Delete a single item by its Id.
        Args:
            id: a string representing the Id of the item to delete.
        Returns: True if the item was deleted, False otherwise.
        """
        if id in self.data:
            del self.data[id]
            await self.save()
            return True
        return False

    async def exists(self, id: str) -> bool:
        return id in self.data

    async def get_by_metadata(self, **metadata) -> list[Any]:
        """
        Get items by metadata key-value pairs.

        Args:
            **metadata: Arbitrary keyword arguments representing metadata key-value pairs to filter items.

        Returns:
            A list of items that match the provided metadata key-value pairs.
        """
        results = []
        for item in self.data.values():
            if all(item.get(k) == v for k, v in metadata.items()):
                results.append(item)
        return results

    async def nearest(self, query: str, top_k: int = 5) -> list[KnwlModel]:
        raise NotImplemented("JsonStorage: semantic search is not available.")

    def __repr__(self):
        return f"<JsonStorage,path={self._path}, save_to_disk={self._save_to_disk}, items={len(self.data)}>"

    def __str__(self):
        return self.__repr__()



import pytest

from knwl.chunking.tiktoken_chunking import TiktokenChunking
from knwl.models.KnwlChunk import KnwlChunk

from knwl.services import services
pytestmark=pytest.mark.llm


@pytest.mark.asyncio
async def test_encode_string():
    chunker = TiktokenChunking()
    content = "Hello, world!"
    tokens = await chunker.encode(content)
    assert isinstance(tokens, list)
    assert all(isinstance(token, int) for token in tokens)


@pytest.mark.asyncio
async def test_decode_tokens():
    chunker = TiktokenChunking()
    content = "Hello, world!"
    tokens = await chunker.encode(content)
    decoded_content = await chunker.decode(tokens)
    assert decoded_content == content


@pytest.mark.asyncio
async def test_chunking():
    chunker = TiktokenChunking(chunk_size=10, chunk_overlap=2)
    content = "This is a test content to be chunked into smaller pieces based on token chunk_size. You can adjust the chunk chunk_size and chunk_overlap as needed."
    chunks = await chunker.chunk(content)
    assert isinstance(chunks, list)
    assert all(isinstance(c, KnwlChunk) for c in chunks)
    assert all(c.tokens > 0 for c in chunks)
    assert all(c.content is not None for c in chunks)
    assert len(chunks) > 1
    assert chunks[0].content in content
    assert chunks[1].content in content
    for chunk in chunks:
        print(f"Chunk (tokens: {chunk.tokens}): {chunk.content}")

@pytest.mark.asyncio
async def test_count_tokens():
    chunker = TiktokenChunking()
    content = "Hello, world!"
    token_count = await chunker.count_tokens(content)
    assert isinstance(token_count, int)
    assert token_count > 0
    assert token_count == len(await chunker.encode(content))

@pytest.mark.asyncio
async def test_decode_tokens_by_tiktoken():
    chunker = TiktokenChunking()
    content = "Hello, world!"
    tokens = await chunker.encode(content)
    decoded_content = await  chunker.decode(tokens)
    assert decoded_content == content


@pytest.mark.asyncio
async def test_chunking_by_token_size():
    chunker = TiktokenChunking(chunk_size=10, chunk_overlap=0)
    content = (
        "This is a test content to be chunked into smaller pieces based on token chunk_size."
    )
    chunks = await chunker.chunk(content)
    assert len(chunks) > 1
    assert chunks[0].content in content
    assert chunks[1].content in content


@pytest.mark.asyncio
async def test_via_services():
    from knwl.config import get_config
    from knwl.chunking.tiktoken_chunking import TiktokenChunking
    chunker = services.create_service("chunking")
    assert chunker.model == get_config("chunking", "tiktoken", "model")
    assert isinstance(chunker, TiktokenChunking)
    assert chunker.chunk_size == 1024
    assert chunker.chunk_overlap == 128

    def create_class_from_dict(name, data):
        return type(name, (), data)

    async def chunk(self, *args, **kwargs):
        return [KnwlChunk(content="special chunk", tokens=10, index=0)]

    SpecialClass = create_class_from_dict("Special", {"name": "Swa", "chunk": chunk})

    config_override = {
        "chunking": {"default": "special", "special": {"class": SpecialClass()}}
    }
    chunker = services.create_service("chunking", override=config_override)
    assert chunker.name == "Swa"
    chunks = await chunker.chunk("Anything")
    assert len(chunks) == 1
    assert chunks[0].content == "special chunk"



import pytest

from knwl.config import get_config
from knwl.services import services
from knwl.storage.chroma_storage import ChromaStorage
from knwl.utils import random_name, get_full_path


@pytest.fixture
def dummy_store():
    storage = ChromaStorage(collection_name="dummy")
    return storage


@pytest.fixture
def dummy_store_with_metadata():
    storage = ChromaStorage(collection_name="dummy", metadata=["a", "b"])
    return storage


@pytest.mark.asyncio
async def test_chroma_db_upsert(dummy_store):
    key = random_name()
    content = key
    data = {key: {"content": content}}
    await dummy_store.upsert(data)
    result = await dummy_store.nearest(key, top_k=1)
    # print(await dummy_store.get_ids())
    # print("querying ", key)
    assert len(result) == 1
    assert result[0]["content"] == content


@pytest.mark.asyncio
async def test_upsert_with_metadata(dummy_store_with_metadata):
    key = random_name()
    content = key
    data = {key: {"content": content, "a": 1, "b": 2}}
    await dummy_store_with_metadata.upsert(data)
    result = await dummy_store_with_metadata.nearest(key, top_k=1)
    assert len(result) == 1
    assert result[0]["content"] == data[key]["content"]


@pytest.mark.asyncio
async def test_query_multiple(dummy_store):
    await dummy_store.clear()
    assert await dummy_store.count() == 0
    data = {"key1": {"content": "data1"}, "key2": {"content": "data2"}}
    await dummy_store.upsert(data)
    result = await dummy_store.nearest("key", top_k=2)
    assert len(result) == 2
    assert {res["content"] for res in result} == {"data1", "data2"}


@pytest.mark.asyncio
async def test_ids(dummy_store):
    await dummy_store.clear()
    data = {"key1": {"content": "data1"}, "key2": {"content": "data2"}}
    await dummy_store.upsert(data)
    ids = await dummy_store.get_ids()
    assert set(ids) == {"key1", "key2"}


@pytest.mark.asyncio
async def test_auto_embedding():
    # chroma does auto-embedding based on all-MiniLM-L6-v2
    from chromadb import Client

    coll = Client()
    collection = coll.get_or_create_collection(name="test_auto_embedding")
    data = {"key1": "This is a test document."}
    collection.upsert(ids=list(data.keys()), documents=list(data.values()))
    all = collection.get(include=["embeddings"])
    assert (
        len(all["embeddings"][0]) == 384
    )  # all-MiniLM-L6-v2 produces 384-dimensional embeddings


@pytest.mark.asyncio
async def test_chroma_via_service():
    chroma = services.create_service("vector")
    assert chroma.path == get_full_path(get_config("vector", "chroma", "path"))
    await chroma.upsert({"a1": {"content": "a1"}})
    names = await chroma.get_collection_names()
    assert chroma.collection_name in names
    config = {
        "vector": {
            "special": {
                "class": "knwl.storage.chroma_storage.ChromaStorage",
                "memory": False,
                "collection_name": "special",
                "path": "$/tests/v2",
            },
        }
    }
    chroma = services.create_service("vector", "special", override=config)
    assert chroma.collection_name == "special"
    await chroma.upsert({"b1": {"content": "b1"}})
    names = await chroma.get_collection_names()
    assert "special" in names
    found = await chroma.get_by_id("b1")
    assert found is not None

@pytest.mark.asyncio
async def test_chroma_memory_service():
    chroma = services.create_service("vector", "memory")
    assert chroma.in_memory is True
    await chroma.upsert({"m1": {"content": "memory data"}})
    found = await chroma.get_by_id("m1")
    assert found is not None
    await chroma.clear()
    found = await chroma.get_by_id("m1")
    assert found is None
    await chroma.upsert({"m2": {"content": "more memory data"}})
    found = await chroma.get_by_id("m2")
    assert found is not None



    


from typing import List

from knwl.logging import log
from knwl.models import (
    KnwlParams,
    KnwlEdge,
    KnwlContext,
    KnwlNode,
    KnwlText,
    KnwlReference,
)
from knwl.models.KnwlInput import KnwlInput
from knwl.semantic.graph_rag.graph_rag_base import GraphRAGBase
from knwl.semantic.graph_rag.strategies.strategy_base import GragStrategyBase
from knwl.utils import unique_strings


class GlobalGragStrategy(GragStrategyBase):

    def __init__(self, grag: GraphRAGBase):
        super().__init__(grag)

    async def augment(self, input: KnwlInput) -> KnwlContext | None:
        """
        The global strategy uses high-level keywords and semantic edge search based on these high-level keywords to find relevant nodes and edges.
        """
        keywords = await self.grag.extract_keywords(input.text)
        if not keywords:
            log.debug("No keywords found for global strategy.")
            return KnwlContext.empty(input)
        if len(keywords.high_level) == 0:
            log.debug("No high level keywords found for global strategy.")
            return KnwlContext.empty(input)
        input.text = ", ".join(
            keywords.high_level
        )  # Override input text with keywords for global topics
        return await self.augment_via_edges(input)



import pytest

from knwl.extraction.basic_keywords_extraction import BasicKeywordsExtraction
from knwl.format import print_knwl
pytestmark = pytest.mark.llm


@pytest.mark.asyncio
async def test_basic():
    input = """Gauss was a child prodigy in mathematics. When the elementary teachers noticed his intellectual abilities, they brought him to the attention of the Duke of Brunswick who sent him to the local Collegium Carolinum,[a] which he attended from 1792 to 1795 with Eberhard August Wilhelm von Zimmermann as one of his teachers. Thereafter the Duke granted him the resources for studies of mathematics, sciences, and classical languages at the University of Göttingen until 1798. His professor in mathematics was Abraham Gotthelf Kästner, whom Gauss called "the leading mathematician among poets, and the leading poet among mathematicians" because of his epigrams.[b] Astronomy was taught by Karl Felix Seyffer, with whom Gauss stayed in correspondence after graduation; Olbers and Gauss mocked him in their correspondence. On the other hand, he thought highly of Georg Christoph Lichtenberg, his teacher of physics, and of Christian Gottlob Heyne, whose lectures in classics Gauss attended with pleasure. Fellow students of this time were Johann Friedrich Benzenberg, Farkas Bolyai, and Heinrich Wilhelm Brandes.."""
    extractor = BasicKeywordsExtraction()
    keywords = await extractor.extract(input)
    assert keywords is not None
    # assert "Gauss" in keywords.low_level_keywords
    # assert "Education" in keywords.high_level_keywords
    print("Have not figured out how to unit test terminal visualization yet.")
    print_knwl(keywords)



import pytest
from faker import Faker


from knwl.llm.anthropic import AnthropicClient
from knwl.models.KnwlAnswer import KnwlAnswer
from knwl.utils import get_full_path
from knwl.services import services

pytestmark = pytest.mark.llm

fake = Faker()


@pytest.mark.asyncio
async def test_basic_ask():

    llm = AnthropicClient()

    # let's change the default caching path
    # note that only the overrides are passed, the rest is taken from default_config
    file_name = fake.word()
    config = {"llm_caching": {"user": {"path": f"$/tests/{file_name}.json"}}}
    llm = services.get_service("llm", "anthropic", override=config)
    assert llm.caching_service is not None
    assert llm.caching_service.path == get_full_path(f"$/tests/{file_name}.json")
    resp = await llm.ask("Hello")
    assert resp is not None
    assert isinstance(resp, KnwlAnswer)

    assert await llm.is_cached("Hello") is True
    file_path = get_full_path(f"$/tests/{file_name}.json")
    import os

    assert os.path.exists(file_path)
    print("")
    print(resp.answer)



from typing import List, Optional
from pydantic import BaseModel, Field, field_validator, model_validator

from knwl.utils import hash_with_prefix


class KnwlAnswer(BaseModel):
    """
    A Pydantic model representing an LLM (Large Language Model) answer with caching capabilities.

    This class encapsulates the response from a LLM along with metadata about
    the query, timing information, and caching status. It automatically generates a unique
    ID based on the messages, service, and model used.

    Attributes:
        messages (list[dict]): The conversation messages exchanged with the LLM.
        llm_model (str): The specific LLM model used (default: "qwen2.5:14b").
        llm_service (str): The LLM service provider (default: "ollama").
        answer (str): The generated response from the LLM.
        timing (float): The time taken to generate the answer in seconds.
        key (str): An optional key for categorizing or identifying the answer.
        category (str): The category or classification of the answer.
        question (str): The original question or prompt that generated this answer.
        from_cache (bool): Indicates whether the answer was retrieved from cache.
        id (Optional[str]): Unique identifier for the answer, auto-generated if not provided.
  
    Example:
        ```python
        llm = services.get_service("llm")
        a = await llm.ask("DNA is the essence of life.")
        print_knwl(a)
        ```
                
╭─────────────────────────────── KnwlLLMAnswer ────────────────────────────────╮
│                                                                              │
│   messages      [1 items]                                                    │
│   llm_model     o14                                                          │
│   llm_service   ollama                                                       │
│   answer        While DNA plays an incredibly important role in biology      │
│                 and genetics, it's more accurate to say that DNA is          │
│                 essential for storing genetic informati...                   │
│   timing        12.5                                                         │
│   key           DNA is the essence of life.                                  │
│   category      none                                                         │
│   question      DNA is the essence of life.                                  │
│   from_cache    True                                                         │
│   id            answer|>6f80fc51a926b69400c1cb5f982074ae                     │
│                                                                              │
╰──────────────────────────────────────────────────────────────────────────────╯

    """
    messages: list[dict] = Field(default_factory=list)
    llm_model: str = Field(default="qwen2.5:14b")
    llm_service: str = Field(default="ollama")
    answer: str = Field(default="")
    timing: float = Field(default=0.0)
    key: str = Field(default="")
    category: str = Field(default="")
    question: str = Field(default="")
    from_cache: bool = Field(default=False)
    id: Optional[str] = Field(default=None)

    @model_validator(mode="after")
    def set_id_if_none(self):
        if self.id is None:
            self.id = KnwlAnswer.hash_keys(
                self.messages, self.llm_service, self.llm_model
            )
        return self

    @staticmethod
    def none() -> "KnwlAnswer":
        return KnwlAnswer(
            answer="I don't know.",
            messages=[],
            llm_model="",
            llm_service="",
            timing=0.0,
            key="",
            category="",
            question="",
            from_cache=False,
        )
    @staticmethod
    def hash_keys(messages: list[dict], llm_service: str, llm_model: str) -> str:
        return hash_with_prefix(str(messages) + llm_service + llm_model, prefix="answer|>")

    def __repr__(self):
        return f"<KnwlLLMAnswer, service={self.llm_service}, model={self.llm_model}, timing={self.timing}, from_cache={self.from_cache}>"

    def __str__(self):
        return self.__repr__()



from typing import List

from knwl.llm.llm_cache_base import LLMCacheBase
from knwl.models import KnwlAnswer
from knwl.storage.json_storage import JsonStorage
from knwl.di import defaults


@defaults("llm_caching", "user")
class JsonLLMCache(LLMCacheBase):
    """
    A thin wrapper around a JSON storage object to provide caching functionality for LLM.
    """

    def __init__(self, path: str = None, enabled: bool = True):
        super().__init__()
        self._path = path
        self.enabled = enabled
        if path is None:
            self.enabled = False

        self.storage = JsonStorage(path=self._path, enabled=self.enabled)

    @property
    def path(self):
        if self.storage is None:
            return None
        return self.storage.path

    async def is_in_cache(
        self, messages: str | list[str] | list[dict], llm_service: str, llm_model: str
    ) -> bool:
        found = await self.get(messages, llm_service, llm_model)
        return found is not None

    async def get(
        self, messages: str | list[str | list[dict]], llm_service: str, llm_model: str
    ) -> KnwlAnswer | None:
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        if isinstance(messages, list):
            messages = [
                {"role": "user", "content": m} if isinstance(m, str) else m
                for m in messages
            ]
        if not messages or len(messages) == 0:
            return None
        key = KnwlAnswer.hash_keys(messages, llm_service, llm_model)
        return await self.get_by_id(key)

    async def get_all_ids(self) -> list[str]:
        return await self.storage.get_all_ids()

    async def save(self):
        if self.enabled:
            await self.storage.save()

    async def clear_cache(self):
        await self.storage.clear_cache()

    async def get_by_id(self, id: str):
        d = await self.storage.get_by_id(id)
        if d is None:
            return None
        return KnwlAnswer(**d)

    async def get_by_ids(self, ids, fields=None):
        return await self.storage.get_by_ids(ids, fields=fields)

    async def filter_new_ids(self, data: list[str]) -> set[str]:
        return await self.storage.filter_new_ids(data)

    async def upsert(self, a: KnwlAnswer):
        if a is None:
            raise ValueError("JsonLLMCache: cannot upsert None in LLMCache.")
        if not isinstance(a, KnwlAnswer):
            raise ValueError("JsonLLMCache: can only upsert KnwlAnswer instances in LLMCache.")
        data = a.model_dump(mode="json")
        data["from_cache"] = True
        blob = {}
        blob[a.id] = data
        await self.storage.upsert(blob)
        await self.save()
        return a.id

    async def delete_by_id(self, id):
        await self.storage.delete_by_id(id)
        await self.save()

    async def delete(self, a: KnwlAnswer):
        await self.storage.delete_by_id(a.id)
        await self.save()

    def __repr__(self):
        return f"<JsonLLMCache, path={self._path}, enabled={self.enabled}>"

    def __str__(self):
        return self.__repr__()



from abc import ABC
from typing import Any
from knwl.framework_base import FrameworkBase
from knwl.models import KnwlModel


class StorageBase(FrameworkBase, ABC):
    """
    Base class for diverse storage implementations.
    This class defines the interface and common properties for storage systems.
    """

    pass

    # todo: turn this into abstract methods
    async def get_by_id(self, id: str) -> KnwlModel | None:
        """
        Retrieves an object by its unique identifier.
        """
        return None

    async def delete_by_id(self, id: str) -> None:
        """
        Deletes an object by its unique identifier.
        """
        pass

    async def exists(self, id: str) -> bool:
        """
        Checks if an object exists by its unique identifier.
        """
        return False

    async def upsert(self, obj: KnwlModel) -> str | None:
        """
        Upserts an object into storage and returns its unique identifier.
        """
        return None

    async def nearest(self, query: str, top_k: int = 5, where: dict[str, Any] | None = None) -> list[KnwlModel]:
        """
        Semantic search, retrieves the top_k nearest objects based on a query.
        """
        return []

    async def get_by_metadata(self, **kwargs) -> list[KnwlModel]:
        """
        Retrieves objects based on metadata key-value pairs.
        Future improvements will define a MongoDB-like query language for more complex queries.
        """
        return []



from typing import Optional, Protocol, runtime_checkable


@runtime_checkable
class KnwlModel(Protocol):
    """
    Protocol defining the interface for all Knwl models.
    
    All Knwl models are Pydantic models with:
    - an `id` field (from hash of key attributes or assigned)
    - `model_dump(mode="json")` for serialization
    """
    
    id: Optional[str]
    
    def model_dump(self, **kwargs) -> dict:
        """Pydantic method for serialization."""
        ...
 